with Strings, Ints;
use Strings, Ints;

pragma Elaborate_All (Strings,Ints);

package body Big_Numbers.Signed.NDRs is

  function Approx(N: Big_Natural) return Flt is
    --- precision not optimal
    Fac: constant Flt := Flt(Limb_Fac);
    L: Unsigned renames N(Extra);
    R: Flt := 0.0;
  begin
    for K in reverse 0 .. L loop
      R := Fac*R+Flt(N(K));
    end loop;
    return R;
  end Approx;

  function Approx(I: Big_Integer) return Flt is
    --- precision not optimal
  begin
    if I.Neg_Sign then return -Approx(I.Number); end if;
    return Approx(I.Number);
  end Approx;

  procedure Show1(St: in String; N: in Big_Natural; NewLine: in Boolean := True) is
  begin
    Show0(St & Strng(N),NewLine);
  end Show1;

  procedure Show1(St: in String; I: in Big_Integer; NewLine: in Boolean := True) is
  begin
    Show0(St & Strng(I),NewLine);
  end Show1;

  procedure Show1(St: in String; S: in NDR; NewLine: in Boolean := True) is
    use Flt_EF;
  begin
    if IsZero(S.Num.Number) then
      Show0(St & "0",NewLine);
    else
      Show1(St & "(",S.Num,False);
      Show1(" / ",S.Den,False);
      Show1(")*Sqrt(",S.Root,False);
      Show1(") =",Sqrt(Approx(S.Root))*Approx(S.Num)/Approx(S.Den),NewLine);
    end if;
  end Show1;

  procedure Put(F: in Byte_File_Type; S: in NDR) is
    LN: constant Natural := Byte_Count(S.Num.Number);
    LD: constant Natural := Byte_Count(S.Den);
    LR: constant Natural := Byte_Count(S.Root);
    LLast: constant Nat3 := Integer_Log(IMax(LN,LD,LR),256);
  begin
    if S.Num.Neg_Sign then Put(F,2*LLast+1,0); else Put(F,2*LLast,0); end if;
    Put(F,S.Num.Number,LLast);
    Put(F,S.Den,LLast);
    Put(F,S.Root,LLast);
  end Put;

  procedure Get(F: in Byte_File_Type; S: out NDR) is
    B: Byte;
    LLast: Nat3;
  begin
    Get(F,B);
    LLast := Nat3(B/2);
    Get(F,S.Num.Number,LLast);
    Get(F,S.Den,LLast);
    Get(F,S.Root,LLast);
    S.Num.Neg_Sign := (B mod 2)=1;
  end Get;

end Big_Numbers.Signed.NDRs;
