
package body Big_Numbers.Signed is

  procedure Assign(I1: in Integer; I2: out Big_Integer) is
  begin
    I2 := (I1<0,(0 => Limb(abs(I1)), others => 0));
  end Assign;

  function IBig(I: Integer) return Big_Integer is
  begin
    return (I<0,(0 => Limb(abs(I)), others => 0));
  end IBig;

  function Sign(I: Big_Integer) return Integer is
  begin
    if I.Number(Extra)=0 and then I.Number(0)=0 then return 0; end if;
    if I.Neg_Sign then return -1; end if;
    return 1;
  end Sign;

  function Compare(I1,I2: Big_Integer) return Integer is
    S1: constant Integer := Sign(I1);
    S2: constant Integer := Sign(I2);
  begin
    if S1>S2 then return  1; end if;
    if S1<S2 then return -1; end if;
    if S1>0 then return  Compare(I1.Number,I2.Number); end if;
    if S1<0 then return -Compare(I1.Number,I2.Number); end if;
    return 0;
  end Compare;

  procedure Neg(Num: in Big_Natural; Neg_Sign: in out Boolean) is
  begin
    if Num(Extra)=0 and then Num(0)=0 then
      Neg_Sign := False;
    else
      Neg_Sign := not Neg_Sign;
    end if;
  end Neg;

  procedure Neg(I: in out Big_Integer) is
  begin
    Neg(I.Number,I.Neg_Sign);
  end Neg;

  function "-"(I: Big_Integer) return Big_Integer is
    It: Big_Integer := I;
  begin
    Neg(It.Number,It.Neg_Sign);
    return It;
  end "-";

  procedure Add(N: in Big_Natural; I: in out Big_Integer) is
    C: Integer;
  begin
    if IsZero(N) then
      null;
    elsif IsZero(I.Number) then
      I := (False,N);
    elsif I.Neg_Sign then
      Basic_Dist(N,I.Number,C);
      I.Neg_Sign := C<0;
    else
      Basic_Add(N,I.Number);
    end if;
  end Add;

  function "+"(N: Big_Natural; I: Big_Integer) return Big_Integer is
    It: Big_Integer := I;
  begin
    Add(N,It);
    return It;
  end "+";

  procedure Add(I1: in Big_Integer; I2: in out Big_Integer) is
  begin
    if I1.Neg_Sign then
      Sub(I1.Number,I2);
    else
      Add(I1.Number,I2);
    end if;
  end Add;

  function "+"(I1,I2: Big_Integer) return Big_Integer is
    It: Big_Integer := I2;
  begin
    Add(I1,It);
    return It;
  end "+";

  procedure Sub(N: in Big_Natural; I: in out Big_Integer) is
    C: Integer;
  begin
    if IsZero(N) then
      null;
    elsif IsZero(I.Number) then
      I := (True,N);
    elsif I.Neg_Sign then
      Basic_Add(N,I.Number);
    else
      Basic_Dist(N,I.Number,C);
      I.Neg_Sign := C>0;
    end if;
  end Sub;

  function "-"(N: Big_Natural; I: Big_Integer) return Big_Integer is
    It: Big_Integer := I;
  begin
    Sub(N,It);
    return It;
  end "-";

  procedure Sub(I1: in Big_Integer; I2: in out Big_Integer) is
  begin
    if I1.Neg_Sign then
      Add(I1.Number,I2);
    else
      Sub(I1.Number,I2);
    end if;
  end Sub;

  function "-"(I1,I2: Big_Integer) return Big_Integer is
    It: Big_Integer := I2;
  begin
    Sub(I1,It);
    return It;
  end "-";

  procedure Mult(I1: in Integer; I2: in out Big_Integer) is
  begin
    Mult(abs(I1),I2.Number);
    if IsZero(I2.Number) then
      I2.Neg_Sign := False;
    elsif I1<0 then
      I2.Neg_Sign := not I2.Neg_Sign;
    end if;
  end Mult;

  function "*"(I1: Integer; I2: Big_Integer) return Big_Integer is
    It: Big_Integer := I2;
  begin
    Mult(I1,It);
    return It;
  end "*";

  procedure Prod(N: in Big_Natural; I1: in Big_Integer; I2: out Big_Integer) is
  begin
    Prod(N,I1.Number,I2.Number);
    if IsZero(I2.Number) then I2.Neg_Sign := False; end if;
  end Prod;

  function "*"(N: Big_Natural; I: Big_Integer) return Big_Integer is
    It: Big_Integer;
  begin
    Prod(N,I,It);
    return It;
  end "*";

  procedure Prod(I1,I2: in Big_Integer; I3: out Big_Integer) is
  begin
    Prod(I1.Number,I2.Number,I3.Number);
    if IsZero(I3.Number) then
      I3.Neg_Sign := False;
    else
      I3.Neg_Sign := I1.Neg_Sign xor I2.Neg_Sign;
    end if;
  end Prod;

  function "*"(I1,I2: Big_Integer) return Big_Integer is
    I3: Big_Integer;
  begin
    Prod(I1,I2,I3);
    return I3;
  end "*";

  procedure Power(J: in Integer; P: in Natural; I: out Big_Integer) is
  begin
    Power(abs(J),P,I.Number);
    if IsZero(I.Number) then
      I.Neg_Sign := False;
    else
      I.Neg_Sign := (J<0) and then (P mod 2)=1;
    end if;
  end Power;

  function "**"(J: Integer; P: Natural) return Big_Integer is
    I: Big_Integer;
  begin
    Power(J,P,I);
    return I;
  end "**";

  procedure Power(I1: in Big_Integer; P: in Natural; I2: out Big_Integer) is
  begin
    Power(I1.Number,P,I2.Number);
    if IsZero(I2.Number) then
      I2.Neg_Sign := False;
    else
      I2.Neg_Sign := I1.Neg_Sign and then (P mod 2)=1;
    end if;
  end Power;

  function "**"(I: Big_Integer; P: Natural) return Big_Integer is
    It: Big_Integer;
  begin
    Power(I,P,It);
    return It;
  end "**";

  procedure Div(J: in Integer; I: in out Big_Integer; R: out Integer) is
  begin
    if J=0 then raise Numeric_Error; end if;
    Div(abs(J),I.Number,R);
    if I.Neg_Sign then R := -R; end if;   -- (A rem B) has the same sign as A, if nonzero
    if IsZero(I.Number) then
      I.Neg_Sign := False;
    else
      I.Neg_Sign := I.Neg_Sign xor (J<0); -- A/B has the same sign as A*B, if nonzero
    end if;
  end Div;

  function "/"(I: Big_Integer; J: Integer) return Big_Integer is
    R: Integer;
    It: Big_Integer := I;
  begin
    Div(J,It,R);
    return It;
  end "/";

  function "rem"(I: Big_Integer; J: Integer) return Integer is
    R: Integer;
    It: Big_Integer := I;
  begin
    Div(J,It,R);
    return R;
  end "rem";

  procedure Div(I1,I2: in Big_Integer; IQ,IR: out Big_Integer) is
  begin
    Div(I1.Number,I2.Number,IQ.Number,IR.Number);
    if IsZero(IR.Number) then
      IR.Neg_Sign := False;
    else
      IR.Neg_Sign := I1.Neg_Sign;                 -- (A rem B) has the same sign as A, if nonzero
    end if;
    if IsZero(IQ.Number) then
      IQ.Neg_Sign := False;
    else
      IQ.Neg_Sign := I1.Neg_Sign xor I2.Neg_Sign; -- A/B has the same sign as A*B, if nonzero
    end if;
  end Div;

  function "/"(I1,I2: Big_Integer) return Big_Integer is
    IQ,IR: Big_Integer;
  begin
    Div(I1,I2,IQ,IR);
    return IQ;
  end "/";

  function "rem"(I1,I2: Big_Integer) return Big_Integer is
    IQ,IR: Big_Integer;
  begin
    Div(I1,I2,IQ,IR);
    return IR;
  end "rem";

  -------------------------------------

  function Strng(I: Big_Integer; Base: Bases := 10) return String is
  begin
    if I.Neg_Sign then return "-" & Strng(I.Number,Base); end if;
    return Strng(I.Number,Base);
  end Strng;

  function Value(S: in String; Base: Bases := 10) return Big_Integer is
    J: Integer := S'First;
    I: Big_Integer;
  begin
    I.Neg_Sign := False;
    loop
      if S(J)=' ' or S(J)='+' then
        null;
      elsif S(J)='-' then
        I.Neg_Sign := not I.Neg_Sign;
      else
        exit;
      end if;
      J := J+1;
    end loop;
    I.Number := Value(S(J .. S'Last),Base);
    if IsZero(I.Number) then I.Neg_Sign := False; end if;
    return I;
  end Value;

end Big_Numbers.Signed;
