
package body Big_Numbers is

  function IsZero(N: Big_Natural) return Boolean is
  begin
    return N(Extra)=0 and then N(0)=0;
  end IsZero;

  procedure Assign(N1: in Natural; N2: out Big_Natural) is
  begin
    N2 := (0 => Limb(N1), others => 0);
  end Assign;

  function NBig(N: Natural) return Big_Natural is
  begin
    return (0 => Limb(N), others => 0);
  end NBig;

  function Compare(N1,N2: Big_Natural) return Integer is
    --- sign of N1-N2
    L1: Unsigned renames N1(Extra);
    L2: Unsigned renames N2(Extra);
  begin
    if L1>L2 then return  1; end if;
    if L1<L2 then return -1; end if;
    for K in reverse 0 .. L1 loop
      if N1(K)>N2(K) then return  1; end if;
      if N1(K)<N2(K) then return -1; end if;
    end loop;
    return 0;
  end Compare;

  procedure Basic_Add(N1: in Big_Natural; N2: in out Big_Natural) is
    --- consider special cases before
    L1: Unsigned renames N1(Extra);
    L2: Unsigned renames N2(Extra);
    Carry: Boolean := False;
    K: Unsigned := 0;
  begin
    loop
      N2(K) := N2(K)+N1(K);
      if Carry then
        Carry := N2(K)<N1(K);
        N2(K) := N2(K)+1;
      else
        Carry := N2(K)<N1(K);
      end if;
      exit when K >= L1 and then not Carry;
      K := K+1;
    end loop;
    if K>L2 then
      L2 := K;
      if L2>Limb_Used then Limb_Used := L2; end if;
    end if;
    while N2(L2)=0 loop            --- normalize
      exit when L2=0;
      L2 := L2-1;
    end loop;
  end Basic_Add;

  procedure Add(N1: in Big_Natural; N2: in out Big_Natural) is
    L1: Unsigned renames N1(Extra);
    L2: Unsigned renames N2(Extra);
  begin
    if N1(L1)=0 then
      if L1 /= 0 then raise Should_Not_Happen; end if;
    elsif N2(L2)=0 then
      if L2 /= 0 then raise Should_Not_Happen; end if;
      N2 := N1;
    else
      Basic_Add(N1,N2);
    end if;
  end Add;

  function "+"(N1,N2: Big_Natural) return Big_Natural is
    L1: Unsigned renames N1(Extra);
    L2: Unsigned renames N2(Extra);
  begin
    if N1(L1)=0 then
      if L1 /= 0 then raise Should_Not_Happen; end if;
      return N2;
    end if;
    if N2(L2)=0 then
      if L2 /= 0 then raise Should_Not_Happen; end if;
      return N1;
    end if;
    declare
      N: Big_Natural := N2;
    begin
      Basic_Add(N1,N);
      return N;
    end;
  end "+";

  procedure Basic_Sub(N1: in Big_Natural; N2: in out Big_Natural) is
    --- do special cases before
    --- assuming N2 >= N1
    L1: Unsigned renames N1(Extra);
    L2: Unsigned renames N2(Extra);
    Carry: Boolean := False;
    N: Limb;
  begin
    for K in Limb_Index loop
      N := N2(K)-N1(K);
      if Carry then
        Carry := N>N2(K);
        N2(K) := N-1;
      else
        Carry := N>N2(K);
        N2(K) := N;
      end if;
      exit when K >= L1 and then not Carry;
    end loop;
    while N2(L2)=0 loop            --- normalize
      exit when L2=0;
      L2 := L2-1;
    end loop;
  end Basic_Sub;

  procedure Basic_NegSub(N1: in Big_Natural; N2: in out Big_Natural) is
    --- do special cases before
    --- assuming N1 >= N2
    L1: Unsigned renames N1(Extra);
    L2: Unsigned renames N2(Extra);
    Carry: Boolean := False;
  begin
    for K in 0 .. L1 loop
      N2(K) := N1(K)-N2(K);
      if Carry then
        Carry := N2(K)>N1(K);
        N2(K) := N2(K)-1;
      else
        Carry := N2(K)>N1(K);
      end if;
    end loop;
    L2 := L1;
    while N2(L2)=0 loop            --- normalize
      exit when L2=0;
      L2 := L2-1;
    end loop;
  end Basic_NegSub;

  procedure Basic_Dist(N1: in Big_Natural; N2: in out Big_Natural; C: out Integer) is
    --- do special cases before
  begin
    C := Compare(N1,N2);
    if C<0 then
      Basic_Sub(N1,N2);
    elsif C>0 then
      Basic_NegSub(N1,N2);
    else
      N2 := Big_NZero;
    end if;
  end Basic_Dist;

  procedure Dist(N1: in Big_Natural; N2: in out Big_Natural) is
    L1: Unsigned renames N1(Extra);
    L2: Unsigned renames N2(Extra);
    C: Integer;
  begin
    if N1(L1)=0 then
      if L1 /= 0 then raise Should_Not_Happen; end if;
    elsif N2(L2)=0 then
      if L2 /= 0 then raise Should_Not_Happen; end if;
      N2 := N1;
    else
      Basic_Dist(N1,N2,C);
    end if;
  end Dist;

  procedure Basic_Mult(P: in Unsigned; N: in out Big_Natural) is
    --- do special cases before
    LP: constant Long := Long(P);
    L: Unsigned renames N(Extra);
    M: Long := 0;
  begin
    for K in 0 .. L loop
      M := LP*Long(N(K))+M;
      N(K) := Limb(M and Limb_Mask);
      M := Shift_Right(M,Limb_Bits);
    end loop;
    if M /= 0 then
      L := L+1;
      if L>Limb_Used then Limb_Used := L; end if;
      N(L) := Limb(M);
    end if;
  end Basic_Mult;

  procedure U_Mult(P: in Unsigned; N: in out Big_Natural) is
    L: Unsigned renames N(Extra);
  begin
    if N(L)=0 then
      if L /= 0 then raise Should_Not_Happen; end if;
    elsif P>1 then
      Basic_Mult(P,N);
    elsif P=0 then
      N := Big_NZero;
    end if;
  end U_Mult;

  procedure Limb_Shift_Left(P: in Unsigned; N: in out Big_Natural) is
    --- multiplication by Limb_Fac**P
    L: Unsigned renames N(Extra);
  begin
    if N(L)=0 then
      if L /= 0 then raise Should_Not_Happen; end if;
    elsif P>0 then
      L := L+P;
      if L>Limb_Used then Limb_Used := L; end if;
      for K in reverse P .. L loop N(K) := N(K-P); end loop;
      for K in reverse 0 .. P-1 loop N(K) := 0; end loop;
    end if;
  end Limb_Shift_Left;

  procedure Shift_Left(P: in Natural; N: in out Big_Natural) is
    --- multiplication by 2**P
  begin
    U_Mult(2**(P mod Limb_Bits),N);
    Limb_Shift_Left(Unsigned(P/Limb_Bits),N);
  end Shift_Left;

  procedure Mult(P: in Natural; N: in out Big_Natural) is
    L: Unsigned renames N(Extra);
  begin
    if N(L)=0 then
      if L /= 0 then raise Should_Not_Happen; end if;
    elsif P>1 then
      Basic_Mult(Unsigned(P),N);
    elsif P=0 then
      N := Big_NZero;
    end if;
  end Mult;

  function "*"(P: Natural; N: Big_Natural) return Big_Natural is
    Nt: Big_Natural := N;
  begin
    Mult(P,Nt);
    return Nt;
  end "*";

  function Simple(N: Big_Natural) return Boolean is
    --- a single nonzero Limb
    L: Unsigned renames N(Extra);
  begin
    if L>0 then
      for K in reverse 0 .. L-1 loop
        if N(K) /= 0 then return False; end if;
      end loop;
    end if;
    return True;
  end Simple;

  procedure Prod(N1,N2: in Big_Natural; N3: out Big_Natural) is
    --- best if N1 is smaller/simpler than N2
    L1: Unsigned renames N1(Extra);
    L2: Unsigned renames N2(Extra);
  begin
    if N1(L1)=0 or else N2(L2)=0 then
      N3 := Big_NZero;
    else
      N3 := N2;
      U_Mult(N1(L1),N3);
      if Simple(N1) then
        Limb_Shift_Left(L1,N3);
      else
        declare
          L3: Unsigned renames N3(Extra);
          procedure ShiftN3 is
          begin
            L3 := L3+1;
            if L3>Limb_Used then Limb_Used := L3; end if;
            for L in reverse 1 .. L3 loop N3(L) := N3(L-1); end loop;
            N3(0) := 0;
          end ShiftN3;
          pragma Inline_Always (ShiftN3);
          Nt: Big_Natural;
        begin
          for K in reverse 0 .. L1-1 loop
            Nt := N2;
            U_Mult(N1(K),Nt);
            ShiftN3;
            Add(Nt,N3);
          end loop;
        end;
      end if;
    end if;
  end Prod;

  function "*"(N1,N2: Big_Natural) return Big_Natural is
    N3: Big_Natural;
  begin
    Prod(N1,N2,N3);
    return N3;
  end "*";

  procedure Basic_Power(M,P: in Unsigned; N: out Big_Natural) is
    --- do special cases before
    Mask: Unsigned := MaxPower2;
    Nt: Big_Natural;
  begin
    while Mask>P loop Mask := Shift_Right(Mask,1); end loop;
    N := (0 => M, others => 0);
    Mask := Shift_Right(Mask,1);
    while Mask /= 0 loop
      Nt := N;
      Prod(Nt,Nt,N);
      if (P and Mask) /= 0 then U_Mult(M,N); end if;
      Mask := Shift_Right(Mask,1);
    end loop;
  end Basic_Power;

  procedure Power(M,P: in Natural; N: out Big_Natural) is
    --- M**P
  begin
    if M=0 then
      if P=0 then raise Numeric_Error; end if;
      N := Big_NZero;
    elsif P=0 then
      N := Big_NOne;
    elsif P=1 then
      N := (0 => Limb(M), others => 0);
    else
      Basic_Power(Unsigned(M),Unsigned(P),N);
    end if;
  end Power;

  function "**"(M,P: Natural) return Big_Natural is
    N: Big_Natural;
  begin
    Power(M,P,N);
    return N;
  end "**";

  procedure Basic_Power(M: in Big_Natural; P: in Unsigned; N: out Big_Natural) is
    --- do special cases before
    Mask: Unsigned := MaxPower2;
    Nt: Big_Natural;
  begin
    while Mask>P loop Mask := Shift_Right(Mask,1); end loop;
    N := M;
    Mask := Shift_Right(Mask,1);
    while Mask /= 0 loop
      Nt := N;
      Prod(Nt,Nt,N);
      if (P and Mask) /= 0 then
        Nt := N;
        Prod(M,Nt,N);
      end if;
      Mask := Shift_Right(Mask,1);
    end loop;
  end Basic_Power;

  procedure Power(M: in Big_Natural; P: in Natural; N: out Big_Natural) is
    --- M**P
  begin
    if IsZero(M) then
      if P=0 then raise Numeric_Error; end if;
      N := Big_NZero;
    elsif P=0 then
      N := Big_NOne;
    else
      N := M;
      if P>1 then
        Basic_Power(M,Unsigned(P),N);
      end if;
    end if;
  end Power;

  function "**"(M: Big_Natural; P: Natural) return Big_Natural is
    N: Big_Natural;
  begin
    Power(M,P,N);
    return N;
  end "**";

  procedure Basic_Div(P: in Unsigned; N: in out Big_Natural; R: out Unsigned) is
    --- do special cases before
    LP: constant Long := Long(P);
    L: Unsigned renames N(Extra);
    Carry: Long := 0;
    M: Long;
  begin
    for K in reverse 0 .. L loop
      M := Limb_Fac*Carry+Long(N(K));
      Carry := M rem LP;
      N(K) := Limb(M/LP);
    end loop;
    while N(L)=0 loop            --- normalize
      exit when L=0;
      L := L-1;
    end loop;
    R := Unsigned(Carry);
  end Basic_Div;

  procedure U_Div(P: in Unsigned; N: in out Big_Natural; R: out Unsigned) is
    --- check before that P /= 0
    L: Unsigned renames N(Extra);
  begin
    if N(L)=0 then
      if L /= 0 then raise Should_Not_Happen; end if;
    elsif P>1 then
      Basic_Div(P,N,R);
    end if;
  end U_Div;

  procedure Limb_Shift_Right(P: in Unsigned; N: in out Big_Natural) is
    --- division by Limb_Fac**P
    L: Unsigned renames N(Extra);
  begin
    if N(L)=0 then
      if L /= 0 then raise Should_Not_Happen; end if;
    elsif P>L then
      L := 0;
      N := (others => 0);
    elsif P>0 then
      L := L-P;
      for K in 0 .. L loop N(K) := N(K+P); end loop;
      for K in L+1 .. L+P loop N(K) := 0; end loop;
    end if;
  end Limb_Shift_Right;

  procedure Shift_Right(P: in Natural; N: in out Big_Natural) is
    --- division by 2**P
    R: Unsigned;
  begin
    Limb_Shift_Right(Unsigned(P/Limb_Bits),N);
    U_Div(2**(P mod Limb_Bits),N,R);
  end Shift_Right;

  procedure Div(P: in Positive; N: in out Big_Natural; R: out Natural) is
    L: Unsigned renames N(Extra);
    UR: Unsigned;
  begin
    if N(L)=0 then
      if L /= 0 then raise Should_Not_Happen; end if;
      UR := 0;
    elsif P>1 then
      Basic_Div(Unsigned(P),N,UR);
    end if;
    R := Natural(UR);
  end Div;

  function "/"(N: Big_Natural; P: Positive) return Big_Natural is
    R: Natural;
    Nt: Big_Natural := N;
  begin
    Div(P,Nt,R);
    return Nt;
  end "/";

  function "rem"(N: Big_Natural; P: Positive) return Natural is
    R: Natural;
    Nt: Big_Natural := N;
  begin
    Div(P,Nt,R);
    return R;
  end "rem";

  procedure Near_Quot(N1,N2: in Big_Natural; NQ: out Big_Natural) is
    --- assuming N2 /= 0
    L1: Unsigned renames N1(Extra);
    L2: Unsigned renames N2(Extra);
    LQ: Unsigned renames NQ(Extra);
  begin
    NQ := Big_NZero;
    if L1<L2 then
      null;
    elsif N1(L1) >= N2(L2) then
      LQ := L1-L2;
      NQ(LQ) := N1(L1)/N2(L2);
    elsif L1=L2 then
      null;
    else
      LQ := L1-L2-1;
      NQ(LQ) := Limb((Limb_Fac*Long(N1(L1))+Long(N1(L1-1)))/Long(N2(L2)));
    end if;
  end Near_Quot;

  procedure Basic_Div(N1: in out Big_Natural; N2: in Big_Natural; NQ: out Big_Natural) is
    --- the remainder of N1/N2 will be in N1
    --- assuming N1 and N2 are nonzero
    Neg: Boolean := False; -- sign of N1
    C: Integer;
    N3,Nt: Big_Natural;
    L3: Unsigned renames N3(Extra);
  begin
    NQ := Big_NZero;
    loop
      Near_Quot(N1,N2,N3);
      exit when N3(L3)=0;
      Prod(N3,N2,Nt);
      if Neg then Basic_Sub(N3,NQ); else Basic_Add(N3,NQ); end if;
      C := Compare(Nt,N1);
      if C<0 then
        Basic_Sub(Nt,N1);
      elsif C>0 then
        Basic_NegSub(Nt,N1);
        Neg := not Neg;
      else
        N1 := Big_NZero;
        Neg := False;
        exit;
      end if;
    end loop;
    if Neg then
      Basic_NegSub(N2,N1);
      Basic_Sub(Big_NOne,NQ);
    end if;
  end Basic_Div;

  function Scaling_Factor(N: Big_Natural) return Unsigned is
    --- assuming N /= 0
    M: Limb renames N(N(Extra));
    Fac: Unsigned := 1;
  begin
    while Fac*M<MaxPower2 loop Fac := 2*Fac; end loop;
    return Fac;
  end Scaling_Factor;

  procedure Div(N: in out Big_Natural; ND: in Big_Natural; NQ: out Big_Natural) is
    -- NQ := N/ND, N becomed remainder
    C: constant Integer := Compare(ND,N);
    LD: Unsigned renames ND(Extra);
  begin
    if ND(LD)=0 then raise Numeric_Error; end if;
    if C=0 then
      NQ := Big_NOne;
      N := Big_NZero;
    elsif C>0 then
      NQ := Big_NZero;
    elsif LD=0 then
      NQ := N;
      N := Big_NZero;
      Basic_Div(ND(0),NQ,N(0));
    else
      declare
        Fac: constant Limb := Scaling_Factor(ND);
        R: Unsigned;
        N2: Big_Natural := ND;
      begin
        U_Mult(Fac,N);
        U_Mult(Fac,N2);
        Basic_Div(N,N2,NQ);
        U_Div(Fac,N,R);
      end;
    end if;
  end Div;

  procedure Div(NN,ND: in Big_Natural; NQ,NR: out Big_Natural) is
  begin
    NR := NN;
    Div(NR,ND,NQ);
  end Div;

  function "/"(N1,N2: Big_Natural) return Big_Natural is
    NR: Big_Natural := N1;
    NQ: Big_Natural;
  begin
    Div(NR,N2,NQ);
    return NQ;
  end "/";

  function "rem"(N1,N2: Big_Natural) return Big_Natural is
    NR: Big_Natural := N1;
    NQ: Big_Natural;
  begin
    Div(NR,N2,NQ);
    return NR;
  end "rem";

  procedure GCD(N,D: in out Unsigned) is
    Tmp: Unsigned;
  begin
    while (N /= 0) loop
      Tmp  := N;
      N := D mod Tmp;
      D := Tmp;
    end loop;
    if D=0 then raise Constraint_Error; end if;
  end GCD;

  procedure GCD(N: in Big_Natural; D: in out Unsigned) is
    --- assuming D /= 0
    R: Unsigned;
    Nt: Big_Natural := N;
  begin
    U_Div(D,Nt,R);
    GCD(R,D);
  end GCD;

  function GCD(N: Big_Natural; P: Positive) return Positive is
    D: Unsigned := Unsigned(P);
  begin
    GCD(N,D);
    return Positive(D);
  end GCD;

  procedure Basic_GCD(N1,N2: in out Big_Natural) is
    --- result is N1
    N3,Nt: Big_Natural;
    L2: Unsigned renames N2(Extra);
  begin
    loop
      if Compare(N1,N2)<0 then
        Nt := N1; N1 := N2; N2 := Nt;
      end if;
      exit when N2(L2)=0;
      Near_Quot(N1,N2,N3);
      Prod(N3,N2,Nt);
      Dist(Nt,N1);
    end loop;
  end Basic_GCD;

  procedure GCD(N1,N2: in Big_Natural; N3: out Big_Natural) is
    L1: Unsigned renames N1(Extra);
    L2: Unsigned renames N2(Extra);
  begin
    if L1=0 then
      if N1(0)=0 then
        if IsZero(N2) then raise Numeric_Error; end if;
        N3 := N2;
      else
        N3 := N1;
        GCD(N2,N3(0));
      end if;
    elsif L2=0 then
      if N2(0)=0 then
        N3 := N1;
      else
        N3 := N2;
        GCD(N1,N3(0));
      end if;
    else
      N3 := N1;
      declare
        N4: Big_Natural := N2;
      begin
        Basic_GCD(N3,N4);
      end;
    end if;
  end GCD;

  procedure DivGCD(N1,N2: in out Big_Natural) is
    N,R,Tmp: Big_Natural;
  begin
    GCD(N1,N2,N);
    Tmp := N1;
    Div(Tmp,N,N1,R);
    if not IsZero(R) then raise Should_Not_Happen; end if;
    Tmp := N2;
    Div(Tmp,N,N2,R);
    if not IsZero(R) then raise Should_Not_Happen; end if;
  end DivGCD;

  -------------------------------------------------------

  function Char_Count(N: Big_Natural; Base: Bases := 10) return Positive is
    R: Natural;
    Nt: Big_Natural := N;
  begin
    for J in 1 .. Positive'Last loop
      Div(Base,Nt,R);
      if IsZero(Nt) then return J; end if;
    end loop;
    raise Should_Not_Happen;
    return 1;
  end Char_Count;

  function Strng(N: Big_Natural; Base: Bases := 10) return String is
    JMax: constant Positive := Char_Count(N,Base);
    R: Natural;
    S: String(1 .. JMax);
    Nt: Big_Natural := N;
  begin
    for J in reverse S'Range loop
      Div(Base,Nt,R);
      S(J) := Dig(R);
    end loop;
    return S;
  end Strng;

  function Value(S: in String; Base: Bases := 10) return Big_Natural is
    VMax: constant Digs := Base-1;
    N,V: Big_Natural := Big_NZero;
  begin
    for J in S'Range loop
      Mult(Base,N);
      V(0) := Limb(Val(S(J),VMax));
      Add(V,N);
    end loop;
    return N;
  end Value;

  function Byte_Count(N: Big_Natural) return Natural is
    L: Unsigned renames N(Extra);
    NL: Limb := N(L);
    J: Natural := Natural(L)*Limb_Bytes;
  begin
    if NL=0 then
      if L /= 0 then raise Should_Not_Happen; end if;
    else
      loop
        NL := Shift_Right(NL,8);
        exit when NL=0;
        J := J+1;
      end loop;
    end if;
    return J;
  end Byte_Count;

  function Convert(N: Big_Natural) return Byte_Vec is
    JMax: constant Natural := Byte_Count(N);
    NL: Limb;
    V: Byte_Vec(0 .. JMax);
  begin
    for J in 0 .. JMax loop
      if (J mod Limb_Bytes)=0 then NL := N(Unsigned(J/Limb_Bytes)); end if;
      V(J) := Byte(NL and 255);
      NL := Shift_Right(NL,8);
    end loop;
    return V;
  end Convert;

  procedure Convert(V: in Byte_Vec; N: out Big_Natural) is
    JMax: constant Integer := V'Last;
    L: Unsigned renames N(Extra);
    J: Natural := V'First;
  begin
    N := Big_NZero;
    loop
      for M in 1 .. Limb_Bytes loop
        if J <= JMax then
          N(L) := N(L)+Limb(V(J));
          J := J+1;
        end if;
        N(L) := Rotate_Right(N(L),8);
      end loop;
      exit when J>JMax;
      L := L+1;
    end loop;
    if L>Limb_Used then Limb_Used := L; end if;
  end Convert;

  procedure Put(F: in Byte_File_Type; N: in Big_Natural; LLast: in Nat3 := Max3) is
    V: constant Byte_Vec := Convert(N);
    L: constant Natural := V'Last-V'First;
  begin
    Put(F,L,LLast);
    Put(F,V);
  end Put;

  procedure Get(F: in Byte_File_Type; N: out Big_Natural; LLast: in Nat3 := Max3) is
    L: Natural;
  begin
    Get(F,L,LLast);
    declare
      V: Byte_Vec(0 .. L);
    begin
      Get(F,V);
      Convert(V,N);
    end;
  end Get;

  function Bits_Used return Positive is
  begin
    return Positive(Limb_Used+1)*Limb_Bits;
  end Bits_Used;

end Big_Numbers;
