with Interfaces, Bytes, Characters, Integer_Log;
use Interfaces, Bytes, Characters;

pragma Elaborate_All (Interfaces,Bytes,Integer_Log);

generic

  Precision: in Positive;

package Big_Numbers is

  type Big_Natural is private;

  function IsZero(N: Big_Natural) return Boolean;
  procedure Assign(N1: in Natural; N2: out Big_Natural);
  function NBig(N: Natural) return Big_Natural;
  function Compare(N1,N2: Big_Natural) return Integer; -- sign of N1-N2
  procedure Add(N1: in Big_Natural; N2: in out Big_Natural);
  function "+"(N1,N2: Big_Natural) return Big_Natural;
  procedure Dist(N1: in Big_Natural; N2: in out Big_Natural);
  procedure Shift_Left(P: in Natural; N: in out Big_Natural);
  procedure Mult(P: in Natural; N: in out Big_Natural);
  function "*"(P: Natural; N: Big_Natural) return Big_Natural;
  procedure Prod(N1,N2: in Big_Natural; N3: out Big_Natural);
  function "*"(N1,N2: Big_Natural) return Big_Natural;
  procedure Power(M,P: in Natural; N: out Big_Natural);
  function "**"(M,P: Natural) return Big_Natural;
  procedure Power(M: in Big_Natural; P: in Natural; N: out Big_Natural);
  function "**"(M: Big_Natural; P: Natural) return Big_Natural;
  procedure Shift_Right(P: in Natural; N: in out Big_Natural);
  procedure Div(P: in Positive; N: in out Big_Natural; R: out Natural);
  function "/"(N: Big_Natural; P: Positive) return Big_Natural;
  function "rem"(N: Big_Natural; P: Positive) return Natural;
  procedure Div(N: in out Big_Natural; ND: in Big_Natural; NQ: out Big_Natural);
  procedure Div(NN,ND: in Big_Natural; NQ,NR: out Big_Natural);
  function "/"(N1,N2: Big_Natural) return Big_Natural;
  function "rem"(N1,N2: Big_Natural) return Big_Natural;
  function GCD(N: Big_Natural; P: Positive) return Positive;
  procedure GCD(N1,N2: in Big_Natural; N3: out Big_Natural);
  procedure DivGCD(N1,N2: in out Big_Natural);

  Max3:  constant Nat3;
  Big_NZero: constant Big_Natural;
  Big_NOne:  constant Big_Natural;

  function Strng(N: Big_Natural; Base: Bases := 10) return String;
  function Value(S: in String; Base: Bases := 10) return Big_Natural;
  function Convert(N: Big_Natural) return Byte_Vec;
  procedure Convert(V: in Byte_Vec; N: out Big_Natural);
  procedure Put(F: in Byte_File_Type; N: in Big_Natural; LLast: in Nat3 := Max3);
  procedure Get(F: in Byte_File_Type; N: out Big_Natural; LLast: in Nat3 := Max3);
  function Bits_Used return Positive;

  pragma Inline_Always (IsZero,Assign,NBig);

private

  subtype Unsigned is Unsigned_32;

  Limb_Bytes: constant Natural := 4;
  Limb_Bits: constant Natural := 8*Limb_Bytes;
  Limbs_NLast: constant Natural := (Precision-1)/Limb_Bits;
  Limbs_ULast: constant Unsigned := Unsigned(Limbs_NLast);
  Extra: constant Unsigned := Limbs_ULast+1;

  subtype Limb is Unsigned;
  subtype Limb_Index is Unsigned range 0 .. Limbs_ULast;
  type Big_Natural is array(0 .. Extra) of Limb;

  procedure Basic_Add(N1: in Big_Natural; N2: in out Big_Natural);
  procedure Basic_Sub(N1: in Big_Natural; N2: in out Big_Natural);
  procedure Basic_NegSub(N1: in Big_Natural; N2: in out Big_Natural);
  procedure Basic_Dist(N1: in Big_Natural; N2: in out Big_Natural; C: out Integer);
  procedure Basic_Mult(P: in Unsigned; N: in out Big_Natural);
  procedure U_Mult(P: in Unsigned; N: in out Big_Natural);
  procedure Limb_Shift_Left(P: in Unsigned; N: in out Big_Natural);
  function Simple(N: Big_Natural) return Boolean;
  procedure Basic_Power(M,P: in Unsigned; N: out Big_Natural);
  procedure Basic_Power(M: in Big_Natural; P: in Unsigned; N: out Big_Natural);
  procedure Basic_Div(P: in Unsigned; N: in out Big_Natural; R: out Unsigned);
  procedure U_Div(P: in Unsigned; N: in out Big_Natural; R: out Unsigned);
  procedure Limb_Shift_Right(P: in Unsigned; N: in out Big_Natural);
  procedure Near_Quot(N1,N2: in Big_Natural; NQ: out Big_Natural);
  procedure Basic_Div(N1: in out Big_Natural; N2: in Big_Natural; NQ: out Big_Natural);
  function Scaling_Factor(N: Big_Natural) return Unsigned;
  procedure GCD(N,D: in out Unsigned);
  procedure GCD(N: in Big_Natural; D: in out Unsigned);
  procedure Basic_GCD(N1,N2: in out Big_Natural);
  function Byte_Count(N: Big_Natural) return Natural;
  function Char_Count(N: Big_Natural; Base: Bases := 10) return Positive;

  pragma Inline_Always (Simple);

  subtype Long is Unsigned_64;
  Long_Bits: constant Positive := 2*Limb_Bits;

  MaxPower2: constant Unsigned := 2**(Limb_Bits-1);
  Limb_Fac:  constant Long := 2**Limb_Bits;
  Limb_Mask: constant Long := Limb_Fac-1;
  Max3: constant Natural := Integer_Log(Limbs_NLast,256);
  Big_NZero: constant Big_Natural := (others => 0);
  Big_NOne: constant Big_Natural := (0 => 1, others => 0);

  Limb_Used: Limb_Index := 0;  --- for range check and statistics
  Should_Not_Happen: exception;

end Big_Numbers;
