with Primes, Ints;
use Primes, Ints;

pragma Elaborate_All (Primes,Ints);

generic

  NMax: in Positive;

package BigRats is

  Size: constant Positive := PrimesLE(NMax);

  subtype BigRat is Int_Vec(1 .. Size);
  type BigRat_Vec is array(Integer range <>) of BigRat;
  subtype Factorial_Type is BigRat_Vec(0 .. NMax);

  function IsOne(B: BigRat) return Boolean;
  function Num(B: BigRat) return BigRat;
  function Den(B: BigRat) return BigRat;
  procedure GCD(B1: in BigRat; B2: in out BigRat);
  procedure GCD(B: in BigRat_Vec; BG: out BigRat);
  procedure Mult(B1: in BigRat; B2: in out BigRat);
  procedure Prod(B1,B2: in BigRat; B3: out BigRat);
  procedure Inv(B: in out BigRat);
  procedure Inv(B1: in BigRat; B2: out BigRat);
  procedure Div(B1: in BigRat; B2: in out BigRat);
  procedure Quot(B1,B2: in BigRat; B3: out BigRat);
  procedure Power(K: in Integer; B: in out BigRat);
  procedure Power(K: in Integer; B1: in BigRat; B2: out BigRat);
  procedure Sqrt(B: in out BigRat; B_Rem: out BigRat);
  procedure DivGCD(B: in out BigRat_Vec; BG: out BigRat);
  procedure Assign(N: in Positive; B: out BigRat);
  procedure Show1(N: in String; B: in BigRat; NewLine: in Boolean := True);
  function Approx(B: BigRat) return Long_Long_Float;
  function Make_Factorials return Factorial_Type;

  pragma Inline_Always (IsOne,Num,Den,Mult,Prod,Inv,Div,Quot,Power,Sqrt);

end BigRats;
