with Globals, Protected_Counters;
use Globals, Protected_Counters;

with Ints;
use Ints;


pragma Elaborate_All(Globals,Protected_Counters);

package body CG.Zerni is

  procedure Initialize_CGS(Parallel: in Boolean := True) is
  begin
    if not Initialized then
      declare
        Task_Error: Boolean := False;
        PC: Protected_Counter(0);

        procedure SSR_Loop is
          L,Js,Jt,Jx: Integer;
          R: Regge;
          Tmp1,Tmp2,Tmp3,Tmp4,Tmp5: Scalar;
        begin
          Proper_Rounding;
          loop
            PC.Next(L);
            exit when L>LMax;
            for S in 0 .. L loop
              Js := L*LMax2+S*LMax1;
              for T in S .. (L+S)/2 loop
                Jt := LST_Count(Js+T)-S+1;
                for X in T .. L loop
                  Jx := Jt+(X-T)*(T-S+1);
                  for B in S .. T loop
                    OrderedRegge(S,B,T,X,L,R);
                    SqrSymRegge(R,C(Jx+B),Tmp1,Tmp2,Tmp3,Tmp4,Tmp5);
                  end loop;
                end loop;
              end loop;
            end loop;
          end loop;
        exception
          when others => Task_Error := True; raise;
        end SSR_Loop;

        task type SSR_Task_Type is end SSR_Task_Type;
        task body SSR_Task_Type is begin SSR_Loop; end SSR_Task_Type;

        NT: constant Natural := Reserve_Tasks(LMax,Parallel);
      begin
        if NT=0 then
          SSR_Loop;
        else
          declare
            SSR_Task: array(1 .. NT) of SSR_Task_Type;
          begin
            SSR_Loop;
          end;
          Free_Tasks(NT);
          if Task_Error then raise Sorry with "Exception in CG.Zerni.Initialize_CGS"; end if;
        end if;
      end;
      Initialized := True;
    end if;
  end Initialize_CGS;

  procedure CGS(I: in Positive; S: in out Scalar) is
  begin
    Copy(C(I),S);
  end CGS;

end CG.Zerni;
