with Strings, Ints, Primes;
use Strings, Ints, Primes;

pragma Elaborate_All (Strings,Ints,Primes);

package body CG is

  procedure Num_Mult(B: in BigRat; S,Tmp3,Tmp4,Tmp5: in out Int_Scalar) is
    BNum: constant BigRat := Num(B);
    SPrime: Int_Scalar renames Tmp3;
    SPower: Int_Scalar renames Tmp4;
    R: Flt;
  begin
    Mult(Two**BNum(1),S,Tmp5);
    for I in 2 .. Size loop
      R := Flt(Prime(I));
      if BNum(I)>1 then
        R := R**BNum(I);
        if R<IntegerLast then -- result is exact
          Mult(R,S,Tmp5);
        else
          Assign(Prime(I),SPrime);
          IPower(BNum(I),SPrime,SPower,Tmp5);
          Mult(SPower,S);
        end if;
      elsif BNum(I)=1 then
        Mult(R,S,Tmp5);
      end if;
    end loop;
  end Num_Mult;

  procedure Den_Mult(B: in BigRat; S,Tmp3,Tmp4,Tmp5: in out Int_Scalar) is
    BDen: constant BigRat := Den(B);
    SPrime: Int_Scalar renames Tmp3;
    SPower: Int_Scalar renames Tmp4;
    R: Flt;
  begin
    Mult(Two**BDen(1),S,Tmp5);
    for I in 2 .. Size loop
      R := Flt(Prime(I));
      if BDen(I)>1 then
        R := R**BDen(I);
        if R<IntegerLast then -- result is exact
          Mult(R,S,Tmp5);
        else
          Assign(Prime(I),SPrime);
          IPower(BDen(I),SPrime,SPower,Tmp5);
          Mult(SPower,S);
        end if;
      elsif BDen(I)=1 then
        Mult(R,S,Tmp5);
      end if;
    end loop;
  end Den_Mult;

  procedure IntSymRegge(R: in Regge; SNum,SDen,SRoot,Tmp3,Tmp4,Tmp5: in out Int_Scalar) is
    --- result is Sqrt(SRoot)*SNum/SDen
    --- R is not being checked for consistency
    JJ: constant Integer := R(I11)+R(I12)+R(I13);
    K1: constant Integer := R(I13);        -- j1+j2-j3
    K2: constant Integer := R(I21);        -- j1-m1
    K3: constant Integer := R(I32);        -- j2+m2
    D1: constant Integer := R(I12)-K2;     -- j3-j2+m1
    D2: constant Integer := R(I11)-K3;     -- j3-j1-m2
    D3: constant Integer := R(I12)-R(I33); -- j1-j2-m3
    KMin: constant Integer := -IMin( 0,D1,D2);
    KMax: constant Integer :=  IMin(K1,K2,K3);
    SK: Scalar renames SDen;
    BS,BRoot: BigRat;
    B: BigRat_Vec(KMin .. KMax);
  begin
    Inv(Factorial(JJ+1),BS);
    for I in R'Range loop
      Mult(Factorial(R(I)),BS);
    end loop;
    Sqrt(BS,BRoot);
    for K in KMin .. KMax loop
      Quot(BS,Factorial(K),B(K));
      Div(Factorial(D1+K),B(K));
      Div(Factorial(D2+K),B(K));
      Div(Factorial(K1-K),B(K));
      Div(Factorial(K2-K),B(K));
      Div(Factorial(K3-K),B(K));
    end loop;
    DivGCD(B,BS);
    Assign(Zero,SNum);
    for K in KMin .. KMax loop
      Assign(One,SK);
      Num_Mult(B(K),SK,Tmp3,Tmp4,Tmp5);
      if (K mod 2)=0 then Add(SK,SNum); else Sub(SK,SNum); end if;
    end loop;
    Num_Mult(BS,SNum,Tmp3,Tmp4,Tmp5);
    Assign(One,SDen);
    Den_Mult(BS,SDen,Tmp3,Tmp4,Tmp5);
    Assign(One,SRoot);
    Num_Mult(BRoot,SRoot,Tmp3,Tmp4,Tmp5);
    if (D3 mod 2) /= 0 then Neg(SNum); end if;
  end IntSymRegge;

  procedure SymRegge(R: in Regge; S,Tmp1,Tmp2,Tmp3,Tmp4,Tmp5: in out Scalar) is
    --- R is not being checked for consistency
    SDen: Scalar renames Tmp1;
    SRoot: Scalar renames Tmp2;
  begin
    IntSymRegge(R,S,SDen,SRoot,Tmp3,Tmp4,Tmp5);
    Div(SDen,S);
    Sqrt(SRoot,Tmp3);
    Mult(Tmp3,S);
  end SymRegge;

  procedure SymRegge(R: in Regge; S: in out Scalar) is
    Tmp1,Tmp2,Tmp3,Tmp4,Tmp5: Scalar;
  begin
    SymRegge(R,S,Tmp1,Tmp2,Tmp3,Tmp4,Tmp5);
  end SymRegge;

  procedure SqrSymRegge(R: in Regge; S,Tmp1,Tmp2,Tmp3,Tmp4,Tmp5: in out Scalar) is
    --- R is not being checked for consistency
    SDen: Scalar renames Tmp1;
    SRoot: Scalar renames Tmp2;
  begin
    IntSymRegge(R,S,SDen,SRoot,Tmp3,Tmp4,Tmp5);
    Div(SDen,S);
    Mult(S,SRoot);
    Mult(SRoot,S);
  end SqrSymRegge;

  procedure SqrSymRegge(R: in Regge; S: in out Scalar) is
    --- R is not being checked for consistency
    Tmp1,Tmp2,Tmp3,Tmp4,Tmp5: Scalar;
  begin
    SqrSymRegge(R,S,Tmp1,Tmp2,Tmp3,Tmp4,Tmp5);
  end SqrSymRegge;

  procedure IntSym3j(N1,N2,N3: in Natural; M1,M2,M3: in Integer; SNum,SDen,SRoot,Tmp3,Tmp4,Tmp5: in out Int_Scalar) is
    --- the arguments are double the usual
    --- result is Sqrt(SRoot)*SNum/SDen
    IsZero: Boolean;
    R: Regge;
  begin
    ToRegge(N1,N2,N3,M1,M2,M3,R,IsZero);
    if IsZero then
      Assign(Zero,SNum);
      Assign(One,SDen);
      Assign(One,SRoot);
    else
      IntSymRegge(R,SNum,SDen,SRoot,Tmp3,Tmp4,Tmp5);
    end if;
  end IntSym3j;

  procedure Sym3j(N1,N2,N3: in Natural; M1,M2,M3: in Integer; S,Tmp1,Tmp2,Tmp3,Tmp4,Tmp5: in out Scalar) is
    --- the arguments are double the usual
    IsZero: Boolean;
    R: Regge;
  begin
    ToRegge(N1,N2,N3,M1,M2,M3,R,IsZero);
    if IsZero then
      Assign(Zero,S);
    else
      SymRegge(R,S,Tmp1,Tmp2,Tmp3,Tmp4,Tmp5);
    end if;
  end Sym3j;

  procedure Sym3j(N1,N2,N3: in Natural; M1,M2,M3: in Integer; S: in out Scalar) is
    --- the arguments are double the usual
    IsZero: Boolean;
    R: Regge;
  begin
    ToRegge(N1,N2,N3,M1,M2,M3,R,IsZero);
    if IsZero then
      Assign(Zero,S);
    else
      SymRegge(R,S);
    end if;
  end Sym3j;

  procedure SqrSym3j(N1,N2,N3: in Natural; M1,M2,M3: in Integer; S,Tmp1,Tmp2,Tmp3,Tmp4,Tmp5: in out Scalar) is
    --- the arguments are double the usual
    IsZero: Boolean;
    R: Regge;
  begin
    ToRegge(N1,N2,N3,M1,M2,M3,R,IsZero);
    if IsZero then
      Assign(Zero,S);
    else
      SqrSymRegge(R,S,Tmp1,Tmp2,Tmp3,Tmp4,Tmp5);
    end if;
  end SqrSym3j;

  procedure SqrSym3j(N1,N2,N3: in Natural; M1,M2,M3: in Integer; S: in out Scalar) is
    --- the arguments are double the usual
    IsZero: Boolean;
    R: Regge;
  begin
    ToRegge(N1,N2,N3,M1,M2,M3,R,IsZero);
    if IsZero then
      Assign(Zero,S);
    else
      SqrSymRegge(R,S);
    end if;
  end SqrSym3j;

  procedure ShowSym3j(N1,N2,N3: in Natural; M1,M2,M3: in Integer) is
    --- the arguments are double the usual
    IsZero: Boolean;
    R: Regge;
  begin
    Show(N1,N2,N3,M1,M2,M3);
    ToRegge(N1,N2,N3,M1,M2,M3,R,IsZero);
    if IsZero then
      Show0("= 0");
    else
      Show(R);
      declare
        SNum,SDen,SRoot,Tmp3,Tmp4,Tmp5: Scalar;
      begin
        IntSymRegge(R,SNum,SDen,SRoot,Tmp3,Tmp4,Tmp5);
        Show0("= ",False);
        Show1("(",SNum,False);
        Show1("/",SDen,False);
        Show1(")*Sqrt(",SRoot,False);
        Show0(")");
        Div(SDen,SNum);
        Sqrt(SRoot,Tmp3);
        Mult(Tmp3,SNum);
        Show1("= ",SNum);
      end;
    end if;
  end ShowSym3j;

  procedure ShowRegges(MaxL: in Natural; Values: in Boolean := False) is
    I: Natural := 0;
    R: Regge;
    Sym,Tmp1,Tmp2,Tmp3,Tmp4,Tmp5: Scalar;
  begin
    for L in 0 .. MaxL loop
      for X in 0 .. L loop
        for T in 0 .. X loop
          for B in 0 .. T loop
            for S in 0 .. B loop
              I := I+1;
              Show1("",I);
              OrderedRegge(S,B,T,X,L,R);
              Show(R);
              if Values then
                SymRegge(R,Sym,Tmp1,Tmp2,Tmp3,Tmp4,Tmp5);
                Show1("",Sym);
              end if;
            end loop;
          end loop;
        end loop;
      end loop;
    end loop;
  end ShowRegges;

end CG;
