with Ints, Flts, Globals, Reps, Params;
use Ints, Flts, Globals, Reps, Params;
with Matrices, Vectors, ScalVectors, CG, CG.Zerni, Zernikes, Zernikes.GFix;

pragma Elaborate_All (Ints,Flts,Globals,Reps,Params);
pragma Elaborate_All (Matrices,Vectors,ScalVectors,CG.Zerni,Zernikes.GFix);

procedure Find_Fixpt is

  NT: constant Positive := GetArg;
  Sol: constant Natural := GetArg;

  P: Param renames PList(Sol);
  Par: Parity renames P.Par;
  Size: Positive renames P.Size;
  Rho: Radius renames P.Rho;

  subtype Scalar is Rep;

  package SCG is new CG (LMax => Size, Scalar => Scalar);
  package SCG_Zerni is new SCG.Zerni;
  use SCG_Zerni;
  package SZ is new Zernikes (Size => Size, Scalar => Scalar);
  package SM is new Matrices (Scalar => Scalar);
  package SV is new Vectors (Component => Scalar);
  package SSV is new ScalVectors (Scalar => Scalar, SM => SM, SV => SV);
  package SZF is new SZ.GFix(Par => Par, Rho => Rho, Mat_PackSize => P.Mat_PackSize,
                             SM => SM, SV => SV, SSV => SSV);
  use SV,SM,SSV,SZ,SZF;

  procedure Show_Eigen(F: in Zernike) is
    Cut: constant Radius := Half;
    Index: Natural := 0;
    EV: Flt;
    B: QC_SimDiag;
    D: Diag renames B.D;
  begin
    DGData(Model(P),F,B);
    for K in D'Range loop
      EV := Approx(D(K));
      if EV >= One then Index := Index+1; end if;
      if EV>Cut then Show1("ev",EV); end if;
    end loop;
    if Par=0 then
      Show1("even index ",Index);
    else
      Show1("odd index ",Index);
    end if;
  end Show_Eigen;

  F: Zernike;
  M: QC_Matrix;

begin
  Show("",P);
  Verbosity := 2;
  PProd_Parallel := True;
  NTasks := NT;
  Initialize_CGS;

  Read(Input_Dir & ApproxFix_Name(P,False),F);
  F.R := Rho;

  if Par=0 then
    --  ApproxFix(Model(P),F,M);
    IterContr(Model(P),F,M);

    Show1("norm:",MaxNorm(F));
    Write(Output_Dir & Mat_Name(P),M);
    Write(Output_Dir & ApproxFix_Name(P,True),F);
  end if;

  Show_Eigen(F);

end Find_Fixpt;
