with Ada.Text_IO, Ada.Finalization, Ada.Real_Time, Tasks, Flts;
use Ada.Text_IO, Ada.Finalization, Ada.Real_Time, Flts;

pragma Elaborate_All (Ada.Text_IO,Ada.Finalization,Ada.Real_Time,Tasks,Flts);

package Messages is

  pragma Elaborate_Body;

  subtype Timer is Time;
  Default_Timer: Timer := Clock;

  function Reset_Timer return Timer renames Clock;
  function Age(T: Timer := Default_Timer) return Flt;
  function Seconds(T: Timer := Default_Timer) return Integer;
  function Minutes(T: Timer := Default_Timer) return Integer;
  function Hours(T: Timer := Default_Timer) return Integer;
  procedure Show_Time(T: in Timer := Default_Timer);             --- H:M:S

  type Tracer is new Controlled with private;

  overriding procedure Adjust(T: in out Tracer);
  overriding procedure Finalize(T: in out Tracer);

  function Starting(Where: String) return Tracer;
  function Seconds(T: Tracer) return Integer;
  function Minutes(T: Tracer) return Integer;
  function Hours(T: Tracer) return Integer;
  procedure Show_Time(T: in Tracer);

private

  package Dur_IO is new Ada.Text_IO.Fixed_IO(Duration);

  type Tracer is new Controlled with
    record
      H: Tasks.Attribute_Handle;       -- task attributes
      T: Timer;
      Wait: Integer;
    end record;

  Adjustements: constant Integer := 2; -- may be compiler dependent

  procedure Show_Time(P: in File_Access; T: in Timer := Default_Timer);

end Messages;
