with Flts, Ints;
use Flts, Ints;

package Params is

  type Model is tagged
    record
      Hill: Flt;
      Ring: Flt;
      F: Natural;
    end record;

  type Param is new Model with
    record
      Size: Positive;
      Par: Parity;
      Symm: Natural;    ----- Symm 1 => odd wrt theta->theta+pi, Symm 2 => odd wrt theta->theta+pi/2
      Rho: Radius;
      Mat_PackSize: Positive;
      Contr_PackSize: Positive;
      NEig: Positive;
      Eig_PackSize: Positive;
      Prefix: String(1..4);
    end record;

  procedure Show(N: in String; P: in Param);
  function Mat_Name(P: Param) return String;
  function ApproxFix_Name(P: Param; Include_Size: Boolean := False) return String;
  function TrueFix_Name(P: Param; Include_Size: Boolean := False) return String;
  function EData_Name(P: Param; Include_Size: Boolean := False) return String;

  HiRe: constant Param := (Hill => Flt(13)*Half**4,
                           Ring => Half**3,
                           F => 0,
                           Size => 140,
                           Par => 0,
                           Symm => 0,
                           Rho => One+Half**6,
                           Mat_PackSize => 20,
                           Contr_PackSize => 100,
                           NEig => 2,              ------ eigenvalues >1
                           Eig_PackSize => 100,
                           Prefix => "HiRi");

  HiRo: constant Param := (Hill => Flt(13)*Half**4,
                           Ring => Half**3,
                           F => 0,
                           Size => 140,
                           Par => 1,
                           Symm => 0,
                           Rho => One+Half**6,
                           Mat_PackSize => 20,
                           Contr_PackSize => 100,
                           NEig => 1,              ------ eigenvalues >=1
                           Eig_PackSize => 100,
                           Prefix => "HiRi");

  He2e: constant Param := (Hill => Zero,
                           Ring => Zero,
                           F => 2,
                           Size => 80,
                           Par => 0,
                           Symm => 0,
                           Rho => One+Half**6,
                           Mat_PackSize => 20,
                           Contr_PackSize => 50,
                           NEig => 1,               ------ eigenvalues >1
                           Eig_PackSize => 60,
                           Prefix => "Hen2");

  He2o: constant Param := (Hill => Zero,           -- two Hen2 bumps
                           Ring => Zero,
                           F => 2,
                           Size => 80,
                           Par => 1,
                           Symm => 0,
                           Rho => One+Half**6,
                           Mat_PackSize => 20,
                           Contr_PackSize => 80,
                           NEig => 1,               ------ eigenvalues >=1
                           Eig_PackSize => 60,
                           Prefix => "Hen2");

  He4e: constant Param := (Hill => Zero,
                           Ring => Zero,
                           F => 4,
                           Size => 70,
                           Par => 0,
                           Symm => 0,
                           Rho => One+Half**6,
                           Mat_PackSize => 20,
                           Contr_PackSize => 70,
                           NEig => 1,               ------ eigenvalues >1
                           Eig_PackSize => 70,
                           Prefix => "Hen4");

  He4o: constant Param := (Hill => Zero,
                           Ring => Zero,
                           F => 4,
                           Size => 70,
                           Par => 1,
                           Symm => 0,
                           Rho => One+Half**6,
                           Mat_PackSize => 20,
                           Contr_PackSize => 70,
                           NEig => 1,               ------ eigenvalues >=1
                           Eig_PackSize => 70,
                           Prefix => "Hen4");

  He6e: constant Param := (Hill => Zero,
                           Ring => Zero,
                           F => 6,
                           Size => 110,
                           Par => 0,
                           Symm => 0,
                           Rho => One+Half**6,
                           Mat_PackSize => 20,
                           Contr_PackSize => 100,
                           NEig => 1,               ------ eigenvalues >1
                           Eig_PackSize => 80,
                           Prefix => "Hen6");

  He6o: constant Param := (Hill => Zero,
                           Ring => Zero,
                           F => 6,
                           Size => 110,
                           Par => 1,
                           Symm => 0,
                           Rho => One+Half**6,
                           Mat_PackSize => 20,
                           Contr_PackSize => 50,
                           NEig => 1,               ------ eigenvalues >=1
                           Eig_PackSize => 80,
                           Prefix => "Hen6");

  H24e: constant Param := (Hill => Zero,           -- four He2 bumps
                           Ring => Zero,
                           F => 2,
                           Size => 140,
                           Par => 0,
                           Symm => 2,
                           Rho => One+Half**6,
                           Mat_PackSize => 25,
                           Contr_PackSize => 140,
                           NEig => 3,               ------ eigenvalues >1
                           Eig_PackSize => 80,
                           Prefix => "H24b");

  H24o: constant Param := (Hill => Zero,           -- four He2 bumps
                           Ring => Zero,
                           F => 2,
                           Size => 140,
                           Par => 1,
                           Symm => 0,
                           Rho => One+Half**6,
                           Mat_PackSize => 25,
                           Contr_PackSize => 140,
                           NEig => 2,               ------ eigenvalues >=1
                           Eig_PackSize => 80,
                           Prefix => "H24b");

  H22e: constant Param := (Hill => Zero,           -- four He2 bumps
                           Ring => Zero,
                           F => 2,
                           Size => 80,
                           Par => 0,
                           Symm => 1,
                           Rho => One+Half**6,
                           Mat_PackSize => 25,
                           Contr_PackSize => 80,
                           NEig => 2,               ------ eigenvalues >1
                           Eig_PackSize => 70,
                           Prefix => "H22b");

  H22o: constant Param := (Hill => Zero,           -- four He2 bumps
                           Ring => Zero,
                           F => 2,
                           Size => 80,
                           Par => 1,
                           Symm => 0,
                           Rho => One+Half**6,
                           Mat_PackSize => 25,
                           Contr_PackSize => 80,
                           NEig => 1,               ------ eigenvalues >=1
                           Eig_PackSize => 70,
                           Prefix => "H22b");

  type Param_List is array(Positive range <>) of Param;

  function Max_Size(P: Param_List) return Positive;

  PList: constant Param_List(1..12) := (HiRe,HiRo,He2e,He2o,He4e,He4o,He6e,He6o,H24e,H24o,H22e,H22o);

end Params;
