with Strings;
use Strings;

pragma Elaborate_All (Strings);

package body Regges is

  procedure ToRegge(N1,N2,N3: in Natural; M1,M2,M3: in Integer; R: out Regge; IsZero: out Boolean) is
    K: Integer;
  begin
    IsZero := True;
    if (M1+M2+M3) /= 0 then return; end if;
    if M1=0 and then M2=0 and then ((N1+N2+N3) mod 4) /= 0 then return; end if;
    --- assign and check
    K := N2+N3-N1; if K<0 or else (K mod 2) /= 0 then return; end if;
    R(I11) := K/2;
    K := N3+N1-N2; if K<0 or else (K mod 2) /= 0 then return; end if;
    R(I12) := K/2;
    K := N1+N2-N3; if K<0 or else (K mod 2) /= 0 then return; end if;
    R(I13) := K/2;
    K := N1-M1;    if K<0 or else (K mod 2) /= 0 then return; end if;
    R(I21) := K/2;
    K := N2-M2;    if K<0 or else (K mod 2) /= 0 then return; end if;
    R(I22) := K/2;
    K := N3-M3;    if K<0 or else (K mod 2) /= 0 then return; end if;
    R(I23) := K/2;
    K := N1+M1;    if K<0 or else (K mod 2) /= 0 then return; end if;
    R(I31) := K/2;
    K := N2+M2;    if K<0 or else (K mod 2) /= 0 then return; end if;
    R(I32) := K/2;
    K := N3+M3;    if K<0 or else (K mod 2) /= 0 then return; end if;
    R(I33) := K/2;
    ---
    IsZero := False;
  end ToRegge;

  procedure To3j(R: in Regge; N1,N2,N3: out Natural; M1,M2,M3: out Integer) is
    --- this is highly non-unique
    Sum: constant Integer := R(I11)+R(I12)+R(I13);
  begin
    N1 := Sum-R(I11);
    N2 := Sum-R(I12);
    N3 := Sum-R(I13);
    M1 := N1-2*R(I21);
    M2 := N2-2*R(I22);
    M3 := N3-2*R(I23);
  end To3j;

  function Check(R: Regge) return Boolean is
    Sum: constant Integer := R(I11)+R(I12)+R(I13);
  begin
    return     (R(I21)+R(I22)+R(I23))=Sum
      and then (R(I31)+R(I32)+R(I33))=Sum
      and then (R(I11)+R(I21)+R(I31))=Sum
      and then (R(I12)+R(I22)+R(I32))=Sum
      and then (R(I13)+R(I23)+R(I33))=Sum;
  end Check;

  procedure Order(R: in out Regge; ChangeSign: out Boolean) is
    --- R is assumed to be consistent
    Row,Col,S,L,K: Integer;
  begin
    ChangeSign := False;
    ------------------------------------------------- find minimum
    Row := 1; Col := 1; S := R(I11);
    if R(I12)<S then           Col := 2; S := R(I12); end if;
    if R(I13)<S then           Col := 3; S := R(I13); end if;
    if R(I21)<S then Row := 2; Col := 1; S := R(I21); end if;
    if R(I22)<S then Row := 2; Col := 2; S := R(I22); end if;
    if R(I23)<S then Row := 2; Col := 3; S := R(I23); end if;
    if R(I31)<S then Row := 3; Col := 1; S := R(I31); end if;
    if R(I32)<S then Row := 3; Col := 2; S := R(I32); end if;
    if R(I33)<S then Row := 3; Col := 3; S := R(I33); end if;
    -------------------------------------------------
    if Row=2 then                                  -- swap rows 1 and 2
      K := R(I11); R(I11) := R(I21); R(I21) := K;
      K := R(I12); R(I12) := R(I22); R(I22) := K;
      K := R(I13); R(I13) := R(I23); R(I23) := K;
      ChangeSign := not ChangeSign;
    elsif Row=3 then                               -- swap rows 1 and 3
      K := R(I11); R(I11) := R(I31); R(I31) := K;
      K := R(I12); R(I12) := R(I32); R(I32) := K;
      K := R(I13); R(I13) := R(I33); R(I33) := K;
      ChangeSign := not ChangeSign;
    end if;
    if Col=2 then                                  -- swap columns 1 and 2
      K := R(I11); R(I11) := R(I12); R(I12) := K;
      K := R(I21); R(I21) := R(I22); R(I22) := K;
      K := R(I31); R(I31) := R(I32); R(I32) := K;
      ChangeSign := not ChangeSign;
    elsif Col=3 then                               -- swap columns 1 and 3
      K := R(I11); R(I11) := R(I13); R(I13) := K;
      K := R(I21); R(I21) := R(I23); R(I23) := K;
      K := R(I31); R(I31) := R(I33); R(I33) := K;
      ChangeSign := not ChangeSign;
    end if;
    ------------------------------------------------- find maximum in 1st row or column
    Row := 1; Col := 1; L := R(I11);
    if R(I12)>L then           Col := 2; L := R(I12); end if;
    if R(I13)>L then           Col := 3; L := R(I13); end if;
    if R(I21)>L then Row := 2; Col := 1; L := R(I21); end if;
    if R(I31)>L then Row := 3; Col := 1; L := R(I31); end if;
    -------------------------------------------------
    if L=S then return; end if;
    if Row /= 1 then                               -- transpose
      K := R(I12); R(I12) := R(I21); R(I21) := K;
      K := R(I13); R(I13) := R(I31); R(I31) := K;
      K := R(I23); R(I23) := R(I32); R(I32) := K;
      K := Row; Row := Col; Col := K;
    end if;
    if Col /= 2 then                               -- swap columns 2 and 3
      K := R(I12); R(I12) := R(I13); R(I13) := K;
      K := R(I22); R(I22) := R(I23); R(I23) := K;
      K := R(I32); R(I32) := R(I33); R(I33) := K;
      ChangeSign := not ChangeSign;
    end if;
    if R(I33)>R(I23) then                          -- swap rows 2 and 3
      K := R(I21); R(I21) := R(I31); R(I31) := K;
      K := R(I22); R(I22) := R(I32); R(I32) := K;
      K := R(I23); R(I23) := R(I33); R(I33) := K;
      ChangeSign := not ChangeSign;
    end if;
    ChangeSign := ChangeSign and then ((R(I11)+R(I12)+R(I13)) mod 2)=1;
  end Order;

  function Indx(S,B,T,X,L: Natural) return Positive is
  begin
    return LST_Count(L*LMax2+S*Lmax1+T)+(X-T)*(T-S+1)+(B-S)+1;
  end Indx;

  procedure CheckIndx is
    I: Natural := 0;
  begin
    for L in 0 .. LMax loop
      for S in 0 .. L loop
        for T in S .. (L+S)/2 loop
          for X in T .. L loop
            for B in S .. T loop
              I := I+1;
              if Indx(S,B,T,X,L) /= I then raise Inconsistent; end if;
            end loop;
          end loop;
        end loop;
      end loop;
    end loop;
    if VSize /= I then raise Inconsistent; end if;
  end CheckIndx;

  procedure Indx(R: in Regge; I: out Positive; ChangeSign: out Boolean) is
    --- R is assumed to be consistent
    Rt: Regge := R;
    S: Integer renames Rt(I11);
    B: Integer renames Rt(I22);
    T: Integer renames Rt(I33);
    X: Integer renames Rt(I21);
    L: Integer renames Rt(I12);
  begin
    Order(Rt,ChangeSign);
    if X<T or T<B then raise Inconsistent; end if;
    I := LST_Count(L*LMax2+S*Lmax1+T)+(X-T)*(T-S+1)+(B-S)+1; -- Indx(S,B,T,X,L)
  end Indx;

  procedure Indx(N1,N2,N3: in Natural; M1,M2,M3: in Integer; I: out Natural; ChangeSign: out Boolean) is
    IsZero: Boolean;
    R: Regge;
  begin
    ToRegge(N1,N2,N3,M1,M2,M3,R,IsZero);
    if IsZero then
      I := 0;
    else
      Order(R,ChangeSign);
      declare
        S: Integer renames R(I11);
        B: Integer renames R(I22);
        T: Integer renames R(I33);
        X: Integer renames R(I21);
        L: Integer renames R(I12);
      begin
        if X<T or T<B then raise Inconsistent; end if;
        I := LST_Count(L*LMax2+S*Lmax1+T)+(X-T)*(T-S+1)+(B-S)+1; -- Indx(S,B,T,X,L)
      end;
    end if;
  end Indx;

  function Indx(N1,N2,N3: Natural; M1,M2,M3: Integer) return Natural is
    IsZero,ChangeSign: Boolean;
    R: Regge;
  begin
    ToRegge(N1,N2,N3,M1,M2,M3,R,IsZero);
    if IsZero then return 0; end if;
    Order(R,ChangeSign);
    declare
      S: Integer renames R(I11);
      B: Integer renames R(I22);
      T: Integer renames R(I33);
      X: Integer renames R(I21);
      L: Integer renames R(I12);
    begin
      if X<T or T<B then raise Inconsistent; end if;
      return LST_Count(L*LMax2+S*Lmax1+T)+(X-T)*(T-S+1)+(B-S)+1; -- Indx(S,B,T,X,L)
    end;
  end Indx;

  procedure OrderedRegge(S,B,T,X,L: in Natural; R: out Regge) is
    --- assuming S <= B <= T <= X <= L, T <= (L+S)/2
  begin
    R := (S,L,X+B-T,X,B,S+L-T,L+B-T,S+X-T,T);
  end OrderedRegge;

  procedure Show(R: in Regge) is
    W: constant Positive := 4;
  begin
    Show0(Strng(R(I11),W) & Strng(R(I12),W) & Strng(R(I13),W));
    Show0(Strng(R(I21),W) & Strng(R(I22),W) & Strng(R(I23),W));
    Show0(Strng(R(I31),W) & Strng(R(I32),W) & Strng(R(I33),W));
  end Show;

  procedure Show(N1,N2,N3: in Natural; M1,M2,M3: in Integer) is
    W: constant Positive := 4;
  begin
    Show0(Strng(N1,W) & "/2" & Strng(N2,W) & "/2" & Strng(N3,W) & "/2" );
    Show0(Strng(M1,W) & "/2" & Strng(M2,W) & "/2" & Strng(M3,W) & "/2" );
  end Show;

  procedure ShowBoth(R: in Regge) is
    N1,N2,N3,M1,M2,M3: Integer;
  begin
    Show(R);
    To3j(R,N1,N2,N3,M1,M2,M3);
    Show(N1,N2,N3,M1,M2,M3);
  end ShowBoth;

end Regges;
