with Strings, Ints, Flts, Globals, Reps, Reps.Balls, Params;
use Strings, Ints, Flts, Globals, Reps.Balls, Params;
with Matrices, Vectors, ScalVectors, CG, CG.Zerni, Zernikes, Zernikes.GFix;

pragma Elaborate_All (Strings,Ints,Flts,Globals,Reps.Balls,Params);
pragma Elaborate_All (Matrices,Vectors,ScalVectors,Zernikes.GFix);

procedure Run_All is

  NT: constant Positive := GetArg;
  Sol: constant Natural := GetArg; -- 0 means: all solutions
  Opt: constant Natural := GetArg; -- 0: fix and index; 1: just fix; 2: just index

  LMax: constant Positive := Max_Size(PList);
  Testing: constant Boolean := False;

  subtype Scalar is Ball;

  package SM is new Matrices (Scalar => Scalar);
  package SV is new Vectors (Component => Scalar);
  package SSV is new ScalVectors (Scalar => Scalar, SM => SM, SV => SV);
  package SCG is new CG (LMax => LMax, Scalar => Scalar);
  package SCG_Zerni is new SCG.Zerni;
  use SCG_Zerni;

  procedure Check_Fix(P: in Param) is
    Size: Positive renames P.Size;
    Par: Parity renames P.Par;
    Rho: Radius renames P.Rho;
    package SZ is new Zernikes (Size => Size, Scalar => Scalar);
    package SZF is new SZ.GFix(Par => Par, Rho => Rho,
                               Mat_PackSize => P.Mat_PackSize,
                               SM => SM, SV => SV, SSV => SSV);
    use SM,SV,SSV,SZ,SZF;
    F0,F: Zernike;
    M: QC_Matrix;
  begin
    Show0("running Check_Fix with");
    Show("",P);
    Read(Input_Dir & ApproxFix_Name(P,False),F0);
    NumTrunc(Size,F0,F);
    CheckSymm(F,P.Symm);
    Read(Input_Dir & Mat_Name(P),M);
    CheckSymm(M,P.Symm);
    F.R := Rho;
    MProd_Parallel := False;
    ContrFix(Model(P),M,P.Contr_PackSize,F);
    CheckNoSymm(F,P.Symm);
    Write(Output_Dir & TrueFix_Name(P,Testing),F);
  end Check_Fix;

  procedure Check_Index(P: in Param) is
    Size: Positive renames P.Size;
    Par: Parity renames P.Par;
    Rho: Radius renames P.Rho;
    NEig: Positive renames P.NEig;
    package SZ is new Zernikes (Size => Size, Scalar => Scalar);
    package SZF is new SZ.GFix(Par => Par, Rho => Rho,
                               Mat_PackSize => P.Mat_PackSize,
                               SM => SM, SV => SV, SSV => SSV);
    use SM,SV,SSV,SZ,SZF;
    S: Scalar;
    F: Zernike;
    D: EData(NEig);
  begin
    Show0("running Check_Index with");
    Show("",P);
    Read(Input_Dir & TrueFix_Name(P,False),F);
    MakeWF2(Model(P),F,D.WFF);
    Partial_Read(Input_Dir & EData_Name(P,False),D);
    Complete(P.Eig_PackSize-2,Model(P),F,D);
    if NEig>1 then
      Show0("KBounds:");
      KBounds(D.WFF,D.U); --- inspect result visually
    end if;
    MProd_Parallel := True;
    LLMapSpec(D,P.Eig_PackSize,S);
    Show0("LLMapSpec:");
    Show1("eigen(" & Strng(Par) & "," & Strng(NEig+1) & ") <=",Sup(S));
    if Sup(S) >= One then raise Sorry with "LLMapSpec failed"; end if;
  end Check_Index;

  procedure Check_Fix(N: in Positive) is
    P: Param renames PList(N);
  begin
    if (Opt /= 2) and (P.Par=0) then
      Check_Fix(P);
    end if;
  end Check_Fix;

  procedure Check_Index(N: in Positive) is
    P: Param renames PList(N);
  begin
    if (Opt /= 1) then
      Check_Index(P);
    end if;
  end Check_Index;

begin

  Verbosity := 1;
  NTasks := NT;
  Show0("computing CG coefficients ...");
  Initialize_CGS;

  If Sol>0 then
    Check_Fix(Sol);
    Check_Index(Sol);
  else
    for N in PList'Range loop
      Check_Fix(N);
      Check_Index(N);
    end loop;
  end if;

end Run_All;
