with Params, Matrices, Vectors, ScalVectors, Linear, Linear.Contr;
use Params;

pragma Elaborate_All (Params,Matrices,Vectors,Linear.Contr);

generic

  Par: in Parity;
  Rho: in Radius;
  Mat_PackSize: in Positive;

  with function Get_Precision(R: Scalar) return Positive is <>;

  with package SM is new Matrices (Scalar => Scalar);
  with package SV is new Vectors (Component => Scalar);
  with package SSV is new ScalVectors (Scalar => Scalar, SM => SM, SV => SV);

package Zernikes.GFix is

  package SV_Ops renames SSV.SV_Ops;
  package Lin is new Linear (Scalar => Scalar, SM => SM, SV => SV, Fun => Zernike, Mode => ZMode, Modes => ZModes);
  package Lin_Contr is new Lin.Contr (SSV => SSV);

  subtype SMatrix is SM.Matrix;
  subtype SVector is SV.Vector;
  subtype SDiag is SSV.Diag;
  subtype SSimDiag is SSV.SimDiag;

  procedure CheckSymm(F: in Zernike; Symm: in Natural);
  procedure CheckNoSymm(F: in Zernike; Symm: in Natural);
  procedure CheckSymm(M: in SMatrix; Symm: in Natural);

  procedure MakeWF2(P: in Model; F: in Zernike; WF,WFF: in out Zernike);
  procedure MakeWF2(P: in Model; F: in Zernike; WFF: in out Zernike);
  function MakeWF2(P: Model; F: Zernike) return Zernike;
  procedure GMap0(P: in Model; F: in Zernike; WFFF,GF: in out Zernike);
  procedure GMap(P: in Model; F: in Zernike; GF: in out Zernike);
  procedure DGMap0(WFF,H: in Zernike; WFFH,DGH: in out Zernike);
  procedure DGMap(WFF,H: in Zernike; DGH: in out Zernike);
  procedure DerMatrix(P: in Model; F: in Zernike; A: in out SMatrix);
  procedure ContrMatrix(P: in Model; F: in Zernike; M: in out SMatrix);
  procedure DContrNorm(P: in Model; M: in SMatrix; F: in Zernike; PackSize: in Positive; S: out Radius);
  procedure ContrFix(P: in Model; M: in SMatrix; PackSize: in Positive; F: in out Zernike);

  --- numerical
  procedure DGData(P: in Model; F: in Zernike; T: in out SSimDiag);
  procedure ApproxFix(P: in Model; F: in out Zernike; M: in out SMatrix; Eps: in Radius := Zero);
  procedure IterContr(P: in Model; F: in out Zernike; M: in out SMatrix);

  QC: constant ZModes := Make(Mat_PackSize,Rho,Par,True,True); --- Simple and no Err modes

  subtype QC_Vector is SVector(QC'Range);
  subtype QC_SimDiag is SSimDiag(QC'Last);
  subtype QC_Matrix is SMatrix(QC'Range,QC'Range);

  ----------------------------------
  ----------------------------------

  type H1Pair is
    record
      H,XH: Zernike; -- normally XH=(-Lap)H
    end record;

  type H1Pairs is array (Natural range <>) of H1Pair;

  procedure IProd(U,V: in H1Pair; S: in out Scalar);
  procedure Copy(U: in H1Pair; V: in out H1Pair);
  procedure AddProd(S: in Scalar; U: in H1Pair; V: in out H1Pair);
  procedure Gram_Schmidt(V: in H1Pairs; U: in out H1Pairs);
  procedure Normalize(U: in out H1Pairs);
  procedure KBounds(WFF: in Zernike; V: in H1Pairs);

  type EData(N: Positive) is
    record
      Last: Natural := N;
      WFF: Zernike;
      Lambda,Fac: SVector(1..N);
      U: H1Pairs(1..N);
    end record;

  procedure Partial_Write(FileName: in String; D: in EData; Decimal: in Boolean := False);
  procedure Partial_Read(FileName: in String; D: in out EData; Decimal: in Boolean := False);
  procedure Complete(Siz: in Positive; P: in Model; F: in Zernike; D: in out EData);
  procedure LLMap0(N: in Natural; D: in EData; H: in Zernike; WH,LH: in out Zernike);
  procedure LLMap(D: in EData; H: in Zernike; LH: in out Zernike);
  procedure LLMapSpec(D: in EData; PackSize: in Positive; S: in out Scalar);

  --- numerical
  procedure Trivial_Eigen(F: in Zernike; D: in out EData);
  procedure Next_Eigen(N: in Positive; D: in out EData; Eps: in Radius := Zero);
  procedure Find_Eigen(F: in Zernike; D: in out EData; Eps: in Radius := Zero);

  DContr_Parallel: Boolean renames Lin.Op_Norm_Parallel;
  DerMatrix_Parallel: Boolean renames Lin.Enclosure_Parallel;

private

  use SM,SV,SSV;

  procedure Complete(Siz: in Positive; U: in out H1Pairs);
  procedure Project(N: in Natural; U: in H1Pairs; H: in out Zernike);
  procedure Bubble_Sort(V: in out FltVec);
  procedure Show_Lower_Bounds(A: in SMatrix);

end Zernikes.GFix;
