with Globals, Strings, Protected_Counters;
use Globals, Strings, Protected_Counters;

pragma Elaborate_All (Globals,Strings,Protected_Counters);

package body Zernikes is

  function FirstE(Last: Natural; E: in Errs) return Natural is
  begin
    for I in 0 .. Last-1 loop
      if E(I) /= Zero then return I; end if;
    end loop;
    return Last;
  end FirstE;

  function LastE(Last: Natural; E: Errs) return Integer is
  begin
    if Not_STrunc then
      for I in reverse 0 .. Last loop
        if E(I) /= Zero then return I; end if;
      end loop;
    end if;
    return -1;
  end LastE;

  function LastC(Last: Natural; C: Poly) return Integer is
    -- returns -1 if C(0..Last) are zero
  begin
    for I in reverse 0 .. Last loop
      if not IsZero(C(I)) then return I; end if;
    end loop;
    return -1;
  end LastC;

  function EffDeg(Last: Natural; C: Poly) return Natural is
  begin
    for I in reverse 1 .. Last loop
      if not IsZero(C(I)) then return I; end if;
    end loop;
    return 0;
  end EffDeg;

  procedure CP(C1: in Poly; C2: in out Poly) is
  begin
    for I in 0 .. PDeg loop Copy(C1(I),C2(I)); end loop;
  end CP;

  -------------------- vector component operations

  function "="(P1,P2: Radial) return Boolean is
    --- equality as sets
    --- assuming same domain
    D: constant Integer := (Size-P1.M)/2;
    C1: Poly renames P1.L.all.Data;
    C2: Poly renames P2.L.all.Data;
  begin
    if P2.M /= P1.M then return False; end if;
    for I in 0 .. D loop
      if C1(I) /= C2(I) then return False; end if;
    end loop;
    if STrunc then return True; end if;
    declare
      E1: Errs renames P1.E;
      E2: Errs renames P2.E;
    begin
      for I in 0 .. D+1 loop
        if E1(I) /= E2(I) then return False; end if;
      end loop;
    end;
    return True;
  end "=";

  function IsZero(P: Radial) return Boolean is
    D: constant Integer := (Size-P.M)/2;
    C: Poly renames P.L.all.Data;
  begin
    for I in 0 .. D loop
      if not IsZero(C(I)) then return False; end if;
    end loop;
    if Not_STrunc then
      declare
        E: Errs renames P.E;
      begin
        for I in 0 .. D+1 loop
          if E(I) /= Zero then return False; end if;
        end loop;
      end;
    end if;
    return True;
  end IsZero;

  function IsReal(P: Radial) return Boolean is
  begin
    if SInfo.IsComplex then
      declare
        D: constant Integer := (Size-P.M)/2;
        C: Poly renames P.L.all.Data;
      begin
        for I in 0 .. D loop
          if not IsReal(C(I)) then return False; end if;
        end loop;
        if Not_STrunc then
          declare
            E: Errs renames P.E;
          begin
            for I in 0 .. D+1 loop
              if E(I) /= Zero then return False; end if;
            end loop;
          end;
        end if;
      end;
    end if;
    return True;
  end IsReal;

  procedure SetZero(P: in out Radial) is
    D: constant Integer := (Size-P.M)/2;
    C: Poly renames P.L.all.Data;
  begin
    for I in 0 .. D loop SetZero(C(I)); end loop;
    if Not_STrunc then
      declare
        E: Errs renames P.E;
      begin
        for I in 0 .. D+1 loop E(I) := Zero; end loop;
      end;
    end if;
  end SetZero;

  procedure Copy(P1: in Radial; P2: in out Radial) is
    D1: constant Integer := (Size-P1.M)/2;
    C1: Poly renames P1.L.all.Data;
    C2: Poly renames P2.L.all.Data;
  begin
    for I in 0 .. D1 loop Copy(C1(I),C2(I)); end loop;
    P2.M := P1.M;
    if Not_STrunc then
      declare
        E1: Errs renames P1.E;
        E2: Errs renames P2.E;
      begin
        for I in 0 .. D1+1 loop E2(I) := E1(I); end loop;
      end;
    end if;
  end Copy;

  procedure Swap(P1,P2: in out Radial) is
    use RP.Alloc;
    L1: constant LD_Pointer := P1.L;
  begin
    if P1.M /= P2.M then raise Undefined; end if;
    P1.L := P2.L; P2.L := L1;
    if Not_STrunc then
      declare
        E1: Errs renames P1.E;
        E2: Errs renames P2.E;
        E: Radius;
      begin
        for I in 0 .. (Size-P1.M)/2+1 loop
          E := E1(I); E1(I) := E2(I); E2(I) := E;
        end loop;
      end;
    end if;
  end Swap;

  procedure Neg(P: in out Radial) is
    C: Poly renames P.L.all.Data;
  begin
    for I in 0 .. (Size-P.M)/2 loop Neg(C(I)); end loop;
  end Neg;

  procedure Neg(P1: in Radial; P2: in out Radial) is
    D1: constant Integer := (Size-P1.M)/2;
    C1: Poly renames P1.L.all.Data;
    C2: Poly renames P2.L.all.Data;
  begin
    for I in 0 .. D1 loop Neg(C1(I),C2(I)); end loop;
    P2.M := P1.M;
    if Not_STrunc then
      declare
        E1: Errs renames P1.E;
        E2: Errs renames P2.E;
      begin
        for I in 0 .. D1+1 loop E2(I) := E1(I); end loop;
      end;
    end if;
  end Neg;

  procedure Add(P1: in Radial; P2: in out Radial) is
    D1: constant Integer := (Size-P1.M)/2;
    C1: Poly renames P1.L.all.Data;
    C2: Poly renames P2.L.all.Data;
  begin
    if P2.M /= P1.M then raise Undefined; end if;
    for I in 0 .. D1 loop Add(C1(I),C2(I)); end loop;
    if Not_STrunc then
      declare
        E1: Errs renames P1.E;
        E2: Errs renames P2.E;
      begin
        for I in 0 .. D1+1 loop E2(I) := E2(I)+E1(I); end loop;
      end;
    end if;
  end Add;

  procedure Sum(P1,P2: in Radial; P3: in out Radial) is
    M1: Natural renames P1.M;
    D1: constant Integer := (Size-M1)/2;
    C1: Poly renames P1.L.all.Data;
    C2: Poly renames P2.L.all.Data;
    C3: Poly renames P3.L.all.Data;
  begin
    if P2.M /= M1 then raise Undefined; end if;
    for I in 0 .. D1 loop Sum(C1(I),C2(I),C3(I)); end loop;
    P3.M := M1;
    if Not_STrunc then
      declare
        E1: Errs renames P1.E;
        E2: Errs renames P2.E;
        E3: Errs renames P3.E;
      begin
        for I in 0 .. D1+1 loop E3(I) := E2(I)+E1(I); end loop;
      end;
    end if;
  end Sum;

  procedure Sub(P1: in Radial; P2: in out Radial) is
    D1: constant Integer := (Size-P1.M)/2;
    C1: Poly renames P1.L.all.Data;
    C2: Poly renames P2.L.all.Data;
  begin
    if P2.M /= P1.M then raise Undefined; end if;
    for I in 0 .. D1 loop Sub(C1(I),C2(I)); end loop;
    if Not_STrunc then
      declare
        E1: Errs renames P1.E;
        E2: Errs renames P2.E;
      begin
        for I in 0 .. D1+1 loop E2(I) := E2(I)+E1(I); end loop;
      end;
    end if;
  end Sub;

  procedure Diff(P1,P2: in Radial; P3: in out Radial) is
    M1: Natural renames P1.M;
    D1: constant Integer := (Size-M1)/2;
    C1: Poly renames P1.L.all.Data;
    C2: Poly renames P2.L.all.Data;
    C3: Poly renames P3.L.all.Data;
  begin
    if P2.M /= M1 then raise Undefined; end if;
    for I in 0 .. D1 loop Diff(C1(I),C2(I),C3(I)); end loop;
    P3.M := M1;
    if Not_STrunc then
      declare
        E1: Errs renames P1.E;
        E2: Errs renames P2.E;
        E3: Errs renames P3.E;
      begin
        for I in 0 .. D1+1 loop E3(I) := E2(I)+E1(I); end loop;
      end;
    end if;
  end Diff;

  procedure Real_Part(P: in out Radial) is
    C: Poly renames P.L.all.Data;
  begin
    if SInfo.IsComplex then
      for I in 0 .. (Size-P.M)/2 loop Real_Part(C(I)); end loop;
    end if;
  end Real_Part;

  procedure Imag_Part(P: in out Radial) is
    C: Poly renames P.L.all.Data;
  begin
    if SInfo.IsComplex then
      for I in 0 .. (Size-P.M)/2 loop Imag_Part(C(I)); end loop;
    else
      SetZero(P);
    end if;
  end Imag_Part;

  procedure Adjoint(P: in out Radial) is
    C: Poly renames P.L.all.Data;
  begin
    if SInfo.IsComplex then
      for I in 0 .. (Size-P.M)/2 loop Adjoint(C(I)); end loop;
    end if;
  end Adjoint;

  procedure Adjoint(P1: in Radial; P2: in out Radial) is
    D1: constant Integer := (Size-P1.M)/2;
    C1: Poly renames P1.L.all.Data;
    C2: Poly renames P2.L.all.Data;
  begin
    for I in 0 .. D1 loop Adjoint(C1(I),C2(I)); end loop;
    P2.M := P1.M;
    if Not_STrunc then
      declare
        E1: Errs renames P1.E;
        E2: Errs renames P2.E;
      begin
        for I in 0 .. D1+1 loop E2(I) := E1(I); end loop;
      end;
    end if;
  end Adjoint;

  procedure Show1(N: in String; P: in Radial; Hide0: in Boolean := True) is
    W: constant Positive := 4;
    D: constant Integer := (Size-P.M)/2;
    C: Poly renames P.L.all.Data;
    AllZero: Boolean;
  begin
    if not IsZero(P) then
      Show1(N & "M ",P.M);
      AllZero := True;
      for I in 0 .. D loop
        if not IsZero(C(I)) then
          Show1(N & "C " & Strng(I,W) & " ",C(I));
          AllZero := False;
        end if;
      end loop;
      if not Hide0 then
        if AllZero then Show0(N & "C 0"); end if;
        Show0;
      end if;
      if Not_STrunc then
        AllZero := True;
        declare
          E: Errs renames P.E;
        begin
          for I in 0 .. D+1 loop
            if E(I) /= Zero then
              Show1(N & "E " & Strng(I,W) & " ",E(I));
              AllZero := False;
            end if;
          end loop;
          if not Hide0 then
            if AllZero then Show0(N & "E 0"); end if;
            Show0;
          end if;
        end;
      end if;
    elsif not Hide0 then
      Show0(N & "0");
    end if;
  end Show1;

  procedure Show2(N: in String; P1,P2: in Radial; Hide0: in Boolean := True) is
    W: constant Positive := 4;
    D1: constant Integer := (Size-P1.M)/2;
    AllZero: Boolean := True;
    C1: Poly renames P1.L.all.Data;
    C2: Poly renames P2.L.all.Data;
  begin
    if P2.M /= P1.M then raise Not_Implemented; end if;
    if not (IsZero(P1) and IsZero(P2)) then
      Show1(N & "M ",P1.M);
      AllZero := True;
      for I in 0 .. D1 loop
        if not (IsZero(C1(I)) and then IsZero(C2(I))) then
          Show2(N & "C " & Strng(I,W) & " ",C1(I),C2(I));
          AllZero := False;
        end if;
      end loop;
      if not Hide0 then
        if AllZero then Show0(N & "C 0 0"); end if;
        Show0;
      end if;
      if Not_STrunc then
        AllZero := True;
        declare
          E1: Errs renames P1.E;
          E2: Errs renames P2.E;
        begin
          for I in 0 .. D1+1 loop
            if (E1(I) /= Zero) or (E2(I) /= Zero) then
              Show2(N & "E " & Strng(I,W) & " ",E1(I),E2(I));
              AllZero := False;
            end if;
          end loop;
          if not Hide0 then
            if AllZero then Show0(N & "E 0 0"); end if;
            Show0;
          end if;
        end;
      end if;
    elsif not Hide0 then
      Show0(N & "0 0");
    end if;
  end Show2;

  procedure Put(F: in File_Type; D: in Natural; E: in RadVec; Decimal: in Boolean := False) is
    L: Natural := D;
  begin
    if STrunc then
      Put(F,0);
      Put(F,Zero,Decimal);
    else
      while L>0 loop
        exit when E(L) /= Zero;
        L := L-1;
      end loop;
      Put(F,L);
      for I in 0 .. L loop Put(F,E(I),Decimal); end loop;
    end if;
  end Put;

  procedure Put(F: in File_Type; P: in Radial; Decimal: in Boolean := False) is
    D: constant Integer := (Size-P.M)/2;
    C: Poly renames P.L.all.Data;
    L: constant Natural := EffDeg(D,C);
  begin
    Put(F,L);
    for I in 0 .. L loop Put(F,C(I),Decimal); end loop;
    Put(F,D+1,P.E,Decimal);
  end Put;

  procedure Get(F: in File_Type; P: in out Radial; Decimal: in Boolean := False) is
    pragma Unreferenced(F,P,Decimal);
  begin
    raise Not_Implemented;
  end Get;

  ---------------------------------- other simple functions

  procedure SetZero(M: in Natural; P: in out Radial) is
  begin
    P.M := M;
    SetZero(P);
  end SetZero;

  function IsSharp(P: Radial) return Boolean is
  begin
    if STrunc then return True; end if;
    declare
      D: constant Integer := (Size-P.M)/2;
      C: Poly renames P.L.all.Data;
      E: Errs renames P.E;
    begin
      for I in 0 .. D loop
        if not IsSharp(C(I)) then return False; end if;
      end loop;
      for I in 0 .. D+1 loop
        if E(I) /= Zero then return False; end if;
      end loop;
    end;
    return True;
  end IsSharp;

  procedure Mult(R: in Flt; P: in out Radial) is
    D: constant Integer := (Size-P.M)/2;
  begin
    if R=Zero then
      SetZero(P);
    else
      declare
        C: Poly renames P.L.all.Data;
        Tmp: Scalar;
      begin
        for I in 0 .. D loop Mult(R,C(I),Tmp); end loop;
      end;
      if Not_STrunc then
        declare
          E: Errs renames P.E;
          A: constant Radius := abs(R);
        begin
          for I in 0 .. D+1 loop E(I) := A*E(I); end loop;
        end;
      end if;
    end if;
  end Mult;

  procedure Prod(R: in Flt; P1: in Radial; P2: in out Radial) is
    D1: constant Integer := (Size-P1.M)/2;
    C1: Poly renames P1.L.all.Data;
    C2: Poly renames P2.L.all.Data;
  begin
    P2.M := P1.M;
    if R=Zero then
      SetZero(P2);
    else
      for I in 0 .. D1 loop Prod(R,C1(I),C2(I)); end loop;
      if Not_STrunc then
        declare
          E1: Errs renames P1.E;
          E2: Errs renames P2.E;
          A: constant Radius := abs(R);
        begin
          for I in 0 .. D1+1 loop E2(I) := A*E1(I); end loop;
        end;
      end if;
    end if;
  end Prod;

  procedure AddProd(R: in Flt; P1: in Radial; P2: in out Radial) is
    D1: constant Integer := (Size-P1.M)/2;
  begin
    if P2.M /= P1.M then raise Undefined; end if;
    if R /= Zero then
      declare
        C1: Poly renames P1.L.all.Data;
        C2: Poly renames P2.L.all.Data;
        Tmp: Scalar;
      begin
        for I in 0 .. D1 loop AddProd(R,C1(I),C2(I),Tmp); end loop;
      end;
      if Not_STrunc then
        declare
          E1: Errs renames P1.E;
          E2: Errs renames P2.E;
          A: constant Radius := abs(R);
        begin
          for I in 0 .. D1+1 loop E2(I) := E2(I)+A*E1(I); end loop;
        end;
      end if;
    end if;
  end AddProd;

  procedure Mult(S: in Scalar; P: in out Radial) is
    D: constant Integer := (Size-P.M)/2;
    R: constant Radius := MaxNorm(S);
    C: Poly renames P.L.all.Data;
  begin
    if R=Zero then
      SetZero(P);
    else
      for I in 0 .. D loop Mult(S,C(I)); end loop;
      if Not_STrunc then
        declare
          E: Errs renames P.E;
        begin
          for I in 0 .. D+1 loop E(I) := R*E(I); end loop;
        end;
      end if;
    end if;
  end Mult;

  procedure Prod(S: in Scalar; P1: in Radial; P2: in out Radial) is
    D1: constant Integer := (Size-P1.M)/2;
    R: constant Radius := MaxNorm(S);
    C1: Poly renames P1.L.all.Data;
    C2: Poly renames P2.L.all.Data;
  begin
    P2.M := P1.M;
    if R=Zero then
      SetZero(P2);
    else
      for I in 0 .. D1 loop Prod(S,C1(I),C2(I)); end loop;
      if Not_STrunc then
        declare
          E1: Errs renames P1.E;
          E2: Errs renames P2.E;
        begin
          for I in 0 .. D1+1 loop E2(I) := R*E1(I); end loop;
        end;
      end if;
    end if;
  end Prod;

  procedure AddProd(S: in Scalar; P1: in Radial; P2: in out Radial) is
    D1: constant Integer := (Size-P1.M)/2;
    R: constant Radius := MaxNorm(S);
  begin
    if P2.M /= P1.M then raise Undefined; end if;
    if R /= Zero then
      declare
        C1: Poly renames P1.L.all.Data;
        C2: Poly renames P2.L.all.Data;
        Tmp: Scalar;
      begin
        for I in 0 .. D1 loop AddProd(S,C1(I),C2(I),Tmp); end loop;
      end;
      if Not_STrunc then
        declare
          E1: Errs renames P1.E;
          E2: Errs renames P2.E;
        begin
          for I in 0 .. D1+1 loop E2(I) := E2(I)+R*E1(I); end loop;
        end;
      end if;
    end if;
  end AddProd;

  procedure SubProd(S: in Scalar; P1: in Radial; P2: in out Radial) is
    D1: constant Integer := (Size-P1.M)/2;
    R: constant Radius := MaxNorm(S);
  begin
    if P2.M /= P1.M then raise Undefined; end if;
    if R /= Zero then
      declare
        C1: Poly renames P1.L.all.Data;
        C2: Poly renames P2.L.all.Data;
        Tmp: Scalar;
      begin
        for I in 0 .. D1 loop SubProd(S,C1(I),C2(I),Tmp); end loop;
      end;
      if Not_STrunc then
        declare
          E1: Errs renames P1.E;
          E2: Errs renames P2.E;
        begin
          for I in 0 .. D1+1 loop E2(I) := E2(I)+R*E1(I); end loop;
        end;
      end if;
    end if;
  end SubProd;

  procedure Div(R: in Flt; P: in out Radial) is
    D: constant Integer := (Size-P.M)/2;
    C: Poly renames P.L.all.Data;
    Tmp: Scalar;
  begin
    for I in 0 .. D loop Div(R,C(I),Tmp); end loop;
    if Not_STrunc then
      declare
        A: constant Radius := One/abs(R);
        E: Errs renames P.E;
      begin
        for I in 0 .. D+1 loop E(I) := A*E(I); end loop;
      end;
    end if;
  end Div;

  procedure Quot(P1: in Radial; R: in Flt; P2: in out Radial) is
    D1: constant Integer := (Size-P1.M)/2;
    C1: Poly renames P1.L.all.Data;
    C2: Poly renames P2.L.all.Data;
  begin
    P2.M := P1.M;
    for I in 0 .. D1 loop Quot(C1(I),R,C2(I)); end loop;
    if Not_STrunc then
      declare
        A: constant Radius := One/abs(R);
        E1: Errs renames P1.E;
        E2: Errs renames P2.E;
      begin
        for I in 0 .. D1+1 loop E2(I) := A*E1(I); end loop;
      end;
    end if;
  end Quot;

  ------------------- more

  function Val(P: Radial; X: Scalar) return Scalar is
    M: Natural renames P.M;
    D: constant Integer := (Size-M)/2;
    S: constant Scalar := Two*Sqr(X)-SOne;
    C: Poly renames P.L.all.Data;
    E: Errs renames P.E;
    N: Integer;
    B: Radius := Zero;
    R0: Scalar;                              -- R^m_{n-2}
    R1: Scalar := SOne;                      -- R^m_n
    R2: Scalar := Flt(M+2)*Sqr(X)-Scal(M+1); -- R^m_{n+2}
    S0,S1,S2,Y: Scalar;
  begin
    if D=0 then
      Y := C(0);
    else
      Y := C(0)*R1+C(1)*R2;     -- N=M
      for I in 1 .. D-1 loop
        N := M+2*I;
        Assign(IRat((N+M)*(N-M),2*N*(N+1)),S0);
        Assign(IRat(M*M,N*(N+2)),S1);
        Sub(S,S1);
        Assign(IRat((N+2+M)*(N+2-M),2*(N+1)*(N+2)),S2);
        Copy(R1,R0);
        Copy(R2,R1);
        R2 := -(S0*R0+S1*R1)/S2;
        Y := Y+C(I+1)*R2;
      end loop;
    end if;
    Y := Y*(X**M);
    if Not_Strunc then
      if not (IsReal(X) and MaxNorm(X)<=One) then raise Not_Implemented; end if;
      for I in reverse 0 .. D+1 loop B := B+E(I); end loop;
      Y := Y+BallAt0(B);     -- valid on the entire domain
    end if;
    return Y;
  end Val;

  function Center0(P: Radial) return Boolean is
    D: constant Integer := (Size-P.M)/2;
    C: Poly renames P.L.all.Data;
  begin
    for I in 0 .. D loop
      if not Center0(C(I)) then return False; end if;
    end loop;
    return True;
  end Center0;

  procedure ErrMult(R: in Radius; P: in out Radial) is
    D: constant Integer := (Size-P.M)/2;
    C: Poly renames P.L.all.Data;
    E: Errs renames P.E;
  begin
    if not STrunc then
      for I in 0 .. D loop ErrMult(R,C(I)); end loop;
      for I in 0 .. D+1 loop E(I) := R*E(I); end loop;
    end if;
  end ErrMult;

  function Contains(P1,P2: Radial) return Logical is
    --- assuming same domain
  begin
    if STrunc then
      if P1=P2 then return True; else return False; end if;
    end if;
    if P2.M /= P1.M then return False; end if;
    declare
      D: constant Integer := (Size-P1.M)/2;
      C1: Poly renames P1.L.all.Data;
      C2: Poly renames P2.L.all.Data;
      E1: Errs renames P1.E;
      E2: Errs renames P2.E;
      Answer: Logical := True;
    begin
      for I in 0 .. D loop
        Answer := Answer and Contains(C1(I),C2(I));
        if Answer=False then return False; end if;
      end loop;
      for I in 0 .. D+1 loop
        if E1(I)<E2(I) then return False; end if;
      end loop;
      return Answer;
    end;
  end Contains;

  procedure NumTrunc(Siz: in Natural; P1: in Radial; P2: in out Radial) is
    M: Natural renames P1.M;
    D1: constant Integer := (Size-M)/2;
    C1: Poly renames P1.L.all.Data;
    C2: Poly renames P2.L.all.Data;
    E2: Errs renames P2.E;
    D: Integer;
  begin
    P2.M := M;
    if Siz >= Size then
      for I in 0 .. D1 loop Center(C1(I),C2(I)); end loop;
    elsif Siz<M then
      for I in 0 .. D1 loop SetZero(C2(I)); end loop;
    else
      D := (Siz-M)/2;
      for I in 0 .. D loop Center(C1(I),C2(I)); end loop;
      for I in D+1 .. D1 loop SetZero(C2(I)); end loop;
    end if;
    if Not_STrunc then
      for I in 0 .. D1+1 loop E2(I) := Zero; end loop;
    end if;
  end NumTrunc;

  function NPower(R: Radius; N: Natural) return Radius is
    --- upper bound on R^N
    I: Integer := N;
    X: Flt := R;
    Y: Flt := One;
  begin
    if N /= 0 then
      loop
        if (I rem 2)=1 then
          Y := X*Y;
        end if;
        I := I/2;
        exit when I=0;
        X := X*X;
      end loop;
    end if;
    return Y;
  end NPower;

  function NPowers(Rho: Radius) return Weights is
    W: Weights;
  begin
    W(0) := One;
    W(1) := Rho;
    for N in 2 .. W'Last loop W(N) := W(N/2)*W(N-N/2); end loop;
    return W;
  end NPowers;

  function NPowersE(Rho: Radius) return WeightsE is
    W: WeightsE;
  begin
    W(0) := One;
    W(1) := Rho;
    for N in 2 .. W'Last loop W(N) := W(N/2)*W(N-N/2); end loop;
    return W;
  end NPowersE;

  procedure Get(Rho: in Radius; F: in File_Type; P: in out Radial; Decimal: in Boolean := False) is
    --- set P.M beforehand
    M: Natural renames P.M;
    C: Poly renames P.L.all.Data;
    E: Errs renames P.E;
    L: Natural := 0;
    DE: Integer := (Size-M)/2+1;
    R: Radius := Zero;
    S: Scalar;
  begin
    SetZero(P);
    if M>Size then DE := 0; end if; -- put norm into P.E(0)
    Get(F,L);
    for I in 0 .. L loop
      Get(F,S,Decimal);
      if I<DE then
        Copy(S,C(I));
      elsif Not_STrunc then
        E(DE) := E(DE)+NPower(Rho,M+2*I)*MaxNorm(S);
      end if;
    end loop;
    Get(F,L);
    for I in 0 .. L loop
      Get(F,R,Decimal);
      if STrunc then
        null;
      elsif I<DE then
        E(I) := E(I)+R;
      else
        E(DE) := E(DE)+R;
      end if;
    end loop;
  end Get;

  procedure Norm(Rho: in Radius; P: in Radial; S: in out Scalar) is
    D: constant Integer := (Size-P.M)/2;
    Rho1: constant Scalar := Scal(Rho);
    Rho2: constant Scalar := Sqr(Rho1);
    C: Poly renames P.L.all.Data;
    T,Tmp: Scalar;
  begin
    Norm(C(D),S);
    for I in reverse 0 .. D-1 loop
      Mult(Rho2,S);
      Norm(C(I),T);
      Add(T,S);
    end loop;
    IPower(P.M,Rho1,T,Tmp);
    Mult(T,S);
    if Not_STrunc then
      declare
        E: Errs renames P.E;
        R: Radius := Zero;
      begin
        for I in reverse 0 .. D+1 loop R := R+E(I); end loop;
        BallAt0(R,T);
        Add(T,S);
      end;
    end if;
  end Norm;

  function MaxNorm(Rho: Radius; P: Radial) return Radius is
    D: constant Integer := (Size-P.M)/2;
    Rho2: constant Radius := Rho*Rho;
    C: Poly renames P.L.all.Data;
    E: Errs renames P.E;
    R: Radius := MaxNorm(C(D));
  begin
    for I in reverse 0 .. D-1 loop
      R := Rho2*R+MaxNorm(C(I));
    end loop;
    R := R*NPower(Rho,P.M);
    if Not_STrunc then
      for I in reverse 0 .. D+1 loop R := R+E(I); end loop;
    end if;
    return R;
  end MaxNorm;

  function MaxNorm(W: RadVec; P: Radial) return Radius is
    --- needs W(2), W(P.M)
    D: constant Integer := (Size-P.M)/2;
    W2: Radius renames W(2);
    C: Poly renames P.L.all.Data;
    E: Errs renames P.E;
    R: Radius := MaxNorm(C(D));
  begin
    for I in reverse 0 .. D-1 loop
      R := W2*R+MaxNorm(C(I));
    end loop;
    R := W(P.M)*R;
    if Not_STrunc then
      for I in reverse 0 .. D+1 loop R := R+E(I); end loop;
    end if;
    return R;
  end MaxNorm;

  procedure L2Prod(SE1,SE2: in Radius; W: in RadVec; P1,P2: in Radial; S: in out Scalar) is
    --- SE1 gets included with P1.E(0), and SE2 gets included with P2.E(0)
    M: Natural renames P1.M;
    D: constant Integer := (Size-M)/2;
    C1: Poly renames P1.L.all.Data;
    C2: Poly renames P2.L.all.Data;
    LC1: constant Integer := LastC(D,C1);           --  -1 if C1=0
    LC2: constant Integer := LastC(D,C2);           --  -1 if C2=0
    N: Natural;
    Tmp,T: Scalar;
  begin
    SetZero(S);
    if P2.M=M then
      N := M;
      for I in 0 .. IMin(LC1,LC2) loop
        if not (IsZero(C1(I)) or else IsZero(C2(I))) then
          Prod(C1(I),C2(I),Tmp);
          Quot(Tmp,Flt(N+1),T);
          Add(T,S);
        end if;
        N := N+2;
      end loop;
      if Not_STrunc then
        declare
          E1: Errs renames P1.E;
          E2: Errs renames P2.E;
          LE1: constant Integer := LastE(D+1,E1);         --  -1 if E1=0
          LE2: constant Integer := LastE(D+1,E2);         --  -1 if E2=0
          N: Integer;
          ES1,ES2: Flt;
          E: Flt := Zero;
        begin
          if (LC1>=0) and (LE2>=0) then -- coeff*err
            N := M;
            ES2 := SE2;
            for I in 0 .. LC1 loop
              ES2 := ES2+E2(I);
              E := E+MaxNorm(C1(I))*ES2*W(N)/Flt(N+1);
              N := N+2;
            end loop;
          end if;
          if (LE1>=0) and (LC2>=0) then -- err*coeff
            N := M;
            ES1 := SE1;
            for I in 0 .. LC2 loop
              ES1 := ES1+E1(I);
              E := E+MaxNorm(C2(I))*ES1*W(N)/Flt(N+1);
              N := N+2;
            end loop;
          end if;
          if (LE1>=0) and (LE2>=0) then -- err*err
            N := M;
            ES1 := SE1;
            ES2 := SE2;
            E := E+(ES1*ES2)*(W(N)*W(N))/Flt(N+1);
            for I in 0 .. IMax(LE1,LE2) loop
              ES1 := ES1+E1(I);
              E := E+(ES1*E2(I)+E1(I)*ES2)*(W(N)*W(N))/Flt(N+1);
              ES2 := ES2+E2(I);
              N := N+2;
            end loop;
          end if;
          BallAt0(E,T);
          Add(T,S);
        end;
      end if;
      Mult(Half,S);
    end if;
  end L2Prod;

  procedure L2Prod(W: in RadVec; P1,P2: in Radial; S: in out Scalar) is
  begin
    L2Prod(Zero,Zero,W,P1,P2,S);
  end L2Prod;

  procedure AddProd(W: in RadVec; P1,P2: in Radial; P3: in out Radial) is
    --- W are norm factors, uses W(0..SizE) but only if Trunc=False
    M1: Natural renames P1.M;
    M2: Natural renames P2.M;
    D2: constant Integer := (Size-M2)/2;
    C2: Poly renames P2.L.all.Data;
    LC2: constant Integer := LastC(D2,C2);             --  -1 if C2=0
    E2: Errs renames P2.E;
    LE2: constant Integer := LastE(D2+1,E2);           --  -1 if E2=0
  begin
    if (P3.M /= M1+M2) and then (P3.M /= abs(M1-M2)) then raise Undefined; end if;
    if (LC2<0) and then (LE2<0) then return; end if;   -- P2=0
    declare
      D1: constant Integer := (Size-M1)/2;
      C1: Poly renames P1.L.all.Data;
      LC1: constant Integer := LastC(D1,C1);           --  -1 if C1=0
      E1: Errs renames P1.E;
      LE1: constant Integer := LastE(D1+1,E1);         --  -1 if E1=0
    begin
      if (LC1<0) and then (LE1<0) then return; end if; -- P1=0
      if (LC1>=0) and then (LC2>=0) then
        declare
          M3: Natural renames P3.M;
          D3: constant Integer := (Size-M3)/2;
          C3: Poly renames P3.L.all.Data;
          E3L: Radius renames P3.E(D3+1);
        begin
          AddProdCC(W,M1,M2,M3,LC1,LC2,C1,C2,C3,E3L);  -- coeff*coeff
        end;
      end if;
      if (LE1<0) and then (LE2<=0) then return; end if;
      if (LC1>=0) and then (LE2>=0) then               -- coeff*err
        AddProdCE(W,M1,M2,P3.M,LC1,LE2,C1,E2,P3.E);
      end if;
      if (LE1>=0) and then (LC2>=0) then               -- err*coeff
        AddProdCE(W,M2,M1,P3.M,LC2,LE1,C2,E1,P3.E);
      end if;
      if (LE1>=0) and then (LE2>=0) then               -- err*err
        declare
          E30: Radius renames P3.E(0);
          B1,B2: Radius := Zero;
        begin
          for I in reverse 0 .. LE1 loop B1 := B1+E1(I); end loop;
          for I in reverse 0 .. LE2 loop B2 := B2+E2(I); end loop;
          E30 := E30+B1*B2;
        end;
      end if;
    end;
  end AddProd;

  procedure IncludeProd(W: in RadVec; P1,P2: in Radial; P3: in out Radial; Subtract: in Boolean) is
  begin
    if Subtract then Neg(P3); end if;
    AddProd(W,P1,P2,P3);
    if Subtract then Neg(P3); end if;
  end IncludeProd;

  procedure Fix_Signs(M1,M2: in out Integer; M3: in Integer) is
    --- if possible, change signs such that M1+M2+M3=0
  begin
    if M3+M2+M1=0 then
      null;
    elsif M3-M2+M1=0 then
      M2 := -M2;
    elsif M3+M2-M1=0 then
      M1 := -M1;
    elsif M3-M2-M1=0 then
      M2 := -M2;
      M1 := -M1;
    else
      raise Undefined with "inconsistent M values";
    end if;
  end Fix_Signs;

  procedure AddProdCC(W: in RadVec; M1,M2,M3,D1,D2: in Natural; C1,C2: in Poly; C3: in out Poly; E3: in out Radius) is
    use Reggez;
    D3: constant Integer := (Size-M3)/2;
    Ms1: Integer := -M1;
    Ms2: Integer := -M2;
    N1,N2,N3,Ind: Integer;
    Coe,RemCoe,C: Scalar;
  begin
    Fix_Signs(Ms1,Ms2,M3);
    for I1 in 0 .. D1 loop
      if not IsZero(C1(I1)) then
        N1 := M1+2*I1;
        for I2 in 0 .. D2 loop
          if not IsZero(C2(I2)) then
            Prod(C1(I1),C2(I2),C);
            Copy(SOne,RemCoe);
            N2 := M2+2*I2;
            for I3 in 0 .. D3 loop
              N3 := M3+2*I3;
              Ind := Indx(N1,N2,N3,Ms1,Ms2,M3);
              if Ind>0 then
                CGS(Ind,Coe);
                Mult(Flt(N3+1),Coe);
                AddProd(Coe,C,C3(I3));
                Sub(Coe,RemCoe);
              end if;
            end loop;
            if Not_STrunc then
              if (N1+N2)>Size then                       -- some terms not included in C3
                E3 := E3+Sup(RemCoe)*W(N1+N2)*MaxNorm(C);
              end if;
            end if;
          end if;
        end loop;
      end if;
    end loop;
  end AddProdCC;

  procedure AddProdCE(W: in RadVec; M1,M2,M3,L1,L2: in Natural; C1: in Poly; E2: in Errs; E3: in out Errs) is
    F2: constant Natural := FirstE(L2,E2);
    L3: constant Integer := (Size-M3)/2+1;
    N1,DN,I3: Integer;
    B1: Radius;
  begin
    for I1 in 0 .. L1 loop
      B1 := MaxNorm(C1(I1));
      if B1 /= Zero then
        N1 := M1+2*I1;
        B1 := W(N1)*B1;                -- annoying factor W, see comment below
        DN := (M2+F2)-N1;
        for I2 in F2 .. L2 loop        -- DN=N2-N1=(M2-M1)+2*(I2-I1)
          I3 := (abs(DN)-M3)/2;
          if I3<0 then I3 := 0; elsif I3>L3 then I3 := L3; end if;
          E3(I3) := E3(I3)+B1*E2(I2);
          DN := DN+2;
        end loop;
      end if;
    end loop;
  end AddProdCE;

  -- only the part of B1 that goes to modes N3>N2 needs to be multiplied by W(N1)
  -- if N1>0 then we could put B1 into two error terms (not useful if Rho is close to 1)

  procedure Raise_Errs(R2,RM: in Radius; Siz: in Natural; P: in out Radial) is
    --- raise errors above Siz
    --- R2=Rho^(-2) and RM=Rho^(-M)
    M: Natural renames P.M;
    D: constant Integer := (Siz-M)/2;
    C: Poly renames P.L.all.Data;
    E: Errs renames P.E;
    Fac: Flt := RM;
    ESum: Flt := Zero;
    S: Scalar;
  begin
    if Not_STrunc then
      if Siz<M then raise Not_Implemented; end if;
      for I in 0 .. D loop
        ESum := ESum+E(I);
        E(I) := Zero;
        BallAt0(ESum*Fac,S);
        Add(S,C(I));
        Fac := Fac*R2;
      end loop;
      E(D+1) := E(D+1)+ESum;
    end if;
  end Raise_Errs;

  procedure InvNegLap(Rho: in Radius; M: in Natural; C1: in Poly; C2: in out Poly; E2: out Radius) is
    D: constant Integer := (Size-M)/2;
    L: constant Integer := LastC(D,C1);
  begin
    for I in 0 .. D loop SetZero(C2(I)); end loop;
    E2 := Zero;
    if L>=0 then
      declare
        N: Natural := M;
        Tmp: Scalar;
      begin
        Quot(C1(0),Flt(4*(M+2)*(M+1)),C2(0));
        Neg(C2(0),C2(1)); -- case D=0 covered after loop
        for I in 1 .. L loop
          N := M+2*I;
          Quot(C1(I),Flt(4*N*(N+1)),Tmp);
          Sub(Tmp,C2(I-1));
          Quot(C1(I),Flt(2*N*(N+2)),Tmp);
          Add(Tmp,C2(I));
          exit when I=D;
          Quot(C1(I),Flt(4*(N+2)*(N+1)),Tmp);
          Sub(Tmp,C2(I+1));
        end loop;
        if Not_STrunc and then L=D then
          E2 := NPower(Rho,N+2)*MaxNorm(C1(D))/Flt(4*(N+2)*(N+1));
        end if;
      end;
    end if;
  end InvNegLap;

  procedure InvNegLap(Rho: in Radius; P1: in Radial; P2: in out Radial) is
    M: Natural renames P1.M;
    DC: constant Integer := (Size-M)/2;
    DE: constant Integer := DC+1;
    C1: Poly renames P1.L.all.Data;
    C2: Poly renames P2.L.all.Data;
    E2: Errs renames P2.E;
    LE: constant Integer := LastE(DE,P1.E);
  begin
    P2.M := M;
    InvNegLap(Rho,M,C1,C2,E2(DE));
    if Not_STrunc then
      for I in 0 .. DC loop E2(I) := Zero; end loop;
    end if;
    if LE>=0 then
      declare
        InvRho: constant Radius := One/Rho;
        Rho2: constant Radius := Rho*Rho;
        InvRho2: constant Radius := InvRho*InvRho;
        E1: Errs := P1.E;
        FE: Integer := FirstE(LE,E1);
        N: Integer;
        Fac0,Fac1,Fac2: Flt;
        Tmp: Scalar;
      begin
        if FE=0 then                      -- distribute err(0) into coeff(0)+err(1)
          Fac2 := One/Flt(4*(M+2)*(M+1));
          BallAt0(NPower(InvRho,M)*Fac2*E1(0),Tmp);
          Add(Tmp,C2(0));
          if DC=0 then
            E2(1) := E2(1)+Rho2*Fac2*E1(0);
          else
            Add(Tmp,C2(1));
          end if;
          E1(1) := E1(1)+E1(0);
          -- E1(0) := Zero;
          FE := 1;
        end if;
        for I in FE .. LE loop
          N := M+2*I;
          Fac0 := InvRho2/Flt(4*N*(N+1));
          Fac1 := One/Flt(2*N*(N+2));
          Fac2 := Rho2/Flt(4*(N+2)*(N+1));
          E2(I-1) := E2(I-1)+Fac0*E1(I);
          E2(I) := E2(I)+Fac1*E1(I);
          exit when I=DE;
          E2(I+1) := E2(I+1)+Fac2*E1(I);
        end loop;
        if LE=DE then
          E2(DE) := E2(DE)+Fac2*E1(DE);
        end if;
      end;
    end if;
  end InvNegLap;

  function InvLapNorm(Rho2: Radius; M: Natural) return Radius is
    N: constant Integer := M+2;
    A: constant Radius := (One/Rho2)/Flt(4*N*(N+1))+One/Flt(2*N*(N+2))+Rho2/Flt(4*(N+2)*(N+1));
    B: constant Radius := (One+Rho2)/Flt(4*(M+2)*(M+1));
  begin
    return RMax(A,B);
  end InvLapNorm;

  procedure Simple_Random(R: in Radius; M: in Natural; P: in out Radial) is
    D: constant Integer := (Size-M)/2;
    Rho: constant Radius := Two*R;
    Rho2: constant Radius := Rho*Rho;
    Fac: Radius := NPower(One/Rho,M);
    C: Poly renames P.L.all.Data;
  begin
    for I in 0 .. D loop
      Simple_Random(C(I));
      Mult(Fac,C(I));
      Fac := Fac/Rho2;
    end loop;
    P.M := M;
    if Not_STrunc then
      declare
        E: Errs renames P.E;
      begin
        for I in 0 .. D+1 loop E(I) := Zero; end loop;
      end;
    end if;
  end Simple_Random;

  ---------------------------- Zernike

  function "="(F1,F2: Zernike) return Boolean is
    --- equality as sets
    C1: Radials renames F1.C;
    C2: Radials renames F2.C;
  begin
    if F1.R /= F2.R then return False; end if;
    if F1.P /= F2.P then return False; end if;
    for M in F1.P .. Size loop
      if C1(M) /= C2(M) then return False; end if;
    end loop;
    if STrunc then return True; end if;
    declare
      E1: Errors renames F1.E;
      E2: Errors renames F2.E;
    begin
      for M in FreqE loop
        if E1(M) /= E2(M) then return False; end if;
      end loop;
    end;
    return True;
  end "=";

  procedure SetZero(F: in out Zernike) is
    -- does not set F.R and F.P
    C: Radials renames F.C;
  begin
    for M in Freq loop SetZero(M,C(M)); end loop;
    if Not_STrunc then F.E := (others => Zero); end if;
  end SetZero;

  procedure SetZero(R: in Radius; F: in out Zernike) is
    -- does not set F.P
    C: Radials renames F.C;
  begin
    F.R := R;
    for M in Freq loop SetZero(M,C(M)); end loop;
    if Not_STrunc then F.E := (others => Zero); end if;
  end SetZero;

  procedure SetZero(R: in Radius; P: in Parity; F: in out Zernike) is
    C: Radials renames F.C;
  begin
    F.R := R;
    F.P := P;
    for M in Freq loop SetZero(M,C(M)); end loop;
    if Not_STrunc then F.E := (others => Zero); end if;
  end SetZero;

  function IsZero(F: Zernike) return Boolean is
    C: Radials renames F.C;
  begin
    for M in F.P .. Size loop
      if not IsZero(C(M)) then return False; end if;
    end loop;
    if Not_STrunc then
      declare
        E: Errors renames F.E;
      begin
        for M in FreqE loop
          if E(M) /= Zero then return False; end if;
        end loop;
      end;
    end if;
    return True;
  end IsZero;

  function IsSharp(F: Zernike) return Boolean is
  begin
    if STrunc then return True; end if;
    declare
      C: Radials renames F.C;
      E: Errors renames F.E;
    begin
      for M in F.P .. Size loop
        if not IsSharp(C(M)) then return False; end if;
      end loop;
      for M in FreqE loop
        if E(M) /= Zero then return False; end if;
      end loop;
    end;
    return True;
  end IsSharp;

  function NonZero(P: Parity; C: Radials) return Boos is
    NZ: Boos;
  begin
    if P=0 then
      NZ(0) := not IsZero(C(0));
    else
      NZ(0) := False;
    end if;
    for M in 1 .. Size loop
      NZ(M) := not IsZero(C(M));
    end loop;
    return NZ;
  end NonZero;

  function NonZero(F: Zernike) return Boos is
    C: Radials renames F.C;
    NZ: Boos;
  begin
    if F.P=0 then
      NZ(0) := not IsZero(C(0));
    else
      NZ(0) := False;
    end if;
    for M in 1 .. Size loop
      NZ(M) := not IsZero(C(M));
    end loop;
    return NZ;
  end NonZero;

  procedure Copy(F1: in Zernike; F2: in out Zernike) is
    C1: Radials renames F1.C;
    C2: Radials renames F2.C;
  begin
    F2.R := F1.R;
    F2.P := F1.P;
    for M in F1.P .. Size loop Copy(C1(M),C2(M)); end loop;
    if Not_STrunc then F2.E := F1.E; end if;
  end Copy;

  procedure Swap(F1,F2: in out Zernike) is
    C1: Radials renames F1.C;
    C2: Radials renames F2.C;
    R: Radius := F1.R;
  begin
    if F1.P /= F2.P then raise Undefined; end if;
    F1.R := F2.R; F2.R := R;
    for M in F1.P .. Size loop Swap(C1(M),C2(M)); end loop;
    if Not_STrunc then
      declare
        E1: Errors renames F1.E;
        E2: Errors renames F2.E;
      begin
        for M in FreqE loop
          R := E1(M); E1(M) := E2(M); E2(M) := R;
        end loop;
      end;
    end if;
  end Swap;

  procedure Neg(F: in out Zernike) is
    C: Radials renames F.C;
  begin
    for M in F.P .. Size loop Neg(C(M)); end loop;
  end Neg;

  procedure Neg(F1: in Zernike; F2: in out Zernike) is
    C1: Radials renames F1.C;
    C2: Radials renames F2.C;
  begin
    F2.R := F1.R;
    F2.P := F1.P;
    for M in F1.P .. Size loop Neg(C1(M),C2(M)); end loop;
    if Not_STrunc then F2.E := F1.E; end if;
  end Neg;

  function "-"(F: Zernike) return Zernike is
    G: Zernike;
  begin
    Neg(F,G);
    return G;
  end "-";

  procedure Add(S: in Scalar; F: in out Zernike) is
  begin
    if F.P=1 then raise Parity_Error; end if;
    Add(S,F.C(0).L.all.Data(0));
  end Add;

  procedure Add(F1: in Zernike; F2: in out Zernike) is
    C1: Radials renames F1.C;
    C2: Radials renames F2.C;
  begin
    if F1.P /= F2.P then raise Parity_Error; end if;
    if F2.R>F1.R then F2.R := F1.R; end if;
    for M in F1.P .. Size loop Add(C1(M),C2(M)); end loop;
    if Not_STrunc then
      declare
        E1: Errors renames F1.E;
        E2: Errors renames F2.E;
      begin
        for M in FreqE loop E2(M) := E2(M)+E1(M); end loop;
      end;
    end if;
  end Add;

  procedure Sum(F1,F2: in Zernike; F3: in out Zernike) is
    C1: Radials renames F1.C;
    C2: Radials renames F2.C;
    C3: Radials renames F3.C;
  begin
    if F1.P /= F2.P then raise Parity_Error; end if;
    F3.P := F1.P;
    F3.R := RMin(F1.R,F2.R);
    for M in F1.P .. Size loop Sum(C1(M),C2(M),C3(M)); end loop;
    if Not_STrunc then
      declare
        E1: Errors renames F1.E;
        E2: Errors renames F2.E;
        E3: Errors renames F3.E;
      begin
        for M in FreqE loop E3(M) := E2(M)+E1(M); end loop;
      end;
    end if;
  end Sum;

  function "+"(F1,F2: Zernike) return Zernike is
    F3: Zernike;
  begin
    Sum(F1,F2,F3);
    return F3;
  end "+";

  procedure Sub(F1: in Zernike; F2: in out Zernike) is
    C1: Radials renames F1.C;
    C2: Radials renames F2.C;
  begin
    if F1.P /= F2.P then raise Parity_Error; end if;
    if F2.R>F1.R then F2.R := F1.R; end if;
    for M in F1.P .. Size loop Sub(C1(M),C2(M)); end loop;
    if Not_STrunc then
      declare
        E1: Errors renames F1.E;
        E2: Errors renames F2.E;
      begin
        for M in FreqE loop E2(M) := E2(M)+E1(M); end loop;
      end;
    end if;
  end Sub;

  procedure Diff(F1,F2: in Zernike; F3: in out Zernike) is
    C1: Radials renames F1.C;
    C2: Radials renames F2.C;
    C3: Radials renames F3.C;
  begin
    if F1.P /= F2.P then raise Parity_Error; end if;
    F3.P := F1.P;
    F3.R := RMin(F1.R,F2.R);
    for M in F1.P .. Size loop Diff(C1(M),C2(M),C3(M)); end loop;
    if Not_STrunc then
      declare
        E1: Errors renames F1.E;
        E2: Errors renames F2.E;
        E3: Errors renames F3.E;
      begin
        for M in FreqE loop E3(M) := E2(M)+E1(M); end loop;
      end;
    end if;
  end Diff;

  function "-"(F1,F2: Zernike) return Zernike is
    F3: Zernike;
  begin
    Diff(F1,F2,F3);
    return F3;
  end "-";

  procedure Mult(S: in Scalar; F: in out Zernike) is
    C: Radials renames F.C;
  begin
    if IsZero(S) then
      SetZero(F);
    else
      for M in F.P .. Size loop Mult(S,C(M)); end loop;
      if Not_STrunc then
        declare
          E: Errors renames F.E;
          R: constant Radius := MaxNorm(S);
        begin
          for M in FreqE loop E(M) := R*E(M); end loop;
        end;
      end if;
    end if;
  end Mult;

  procedure Prod(S: in Scalar; F1: in Zernike; F2: in out Zernike) is
    C1: Radials renames F1.C;
    C2: Radials renames F2.C;
  begin
    F2.P := F1.P;
    F2.R := F1.R;
    if IsZero(S) then
      SetZero(F2);
    else
      for M in F1.P .. Size loop Prod(S,C1(M),C2(M)); end loop;
      if Not_STrunc then
        declare
          E1: Errors renames F1.E;
          E2: Errors renames F2.E;
          R: constant Radius := MaxNorm(S);
        begin
          for M in FreqE loop E2(M) := R*E1(M); end loop;
        end;
      end if;
    end if;
  end Prod;

  function "*"(S: Scalar; F: Zernike) return Zernike is
    G: Zernike;
  begin
    Prod(S,F,G);
    return G;
  end "*";

  procedure AddProd(S: in Scalar; F1: in Zernike; F2: in out Zernike) is
  begin
    if not IsZero(S) then
      if F1.P /= F2.P then raise Parity_Error; end if;
      if F2.R>F1.R then F2.R := F1.R; end if;
      declare
        C1: Radials renames F1.C;
        C2: Radials renames F2.C;
      begin
        for M in F1.P .. Size loop AddProd(S,C1(M),C2(M)); end loop;
      end;
      if Not_STrunc then
        declare
          E1: Errors renames F1.E;
          E2: Errors renames F2.E;
          R: constant Radius := MaxNorm(S);
        begin
          for M in FreqE loop E2(M) := E2(M)+R*E1(M); end loop;
        end;
      end if;
    end if;
  end AddProd;

  procedure Mult(F1: in Zernike; F2: in out Zernike) is
    F3: Zernike;
  begin
    Prod(F1,F2,F3);
    Swap(F2,F3);
  end Mult;

  procedure Oddify(E: in out Errors) is
  begin
    if E(0) /= Zero then
      E(1) := E(1)+E(0);
      E(0) := Zero;
    end if;
  end Oddify;

  function LastE(E: Errors) return Integer is
  begin
    if Not_Strunc then
      for M in reverse FreqE loop
        if E(M) /= Zero then return M; end if;
      end loop;
    end if;
    return -1;
  end LastE;

  function LastC(First: Parity; C: Radials) return Integer is
    --- returns -1 if C(First..Size) are zero
  begin
    for M in reverse First .. Size loop
      if not IsZero(C(M)) then return M; end if;
    end loop;
    return -1;
  end LastC;

  function ToErr(W: RadVec; NZ: Boos; F: Zernike) return Weights is
    --- assuming NZ=NonZero(F)
    E: Weights;
    C: Radials renames F.C;
  begin
    for M in Freq loop
      if NZ(M) then E(M) := MaxNorm(W,C(M)); else E(M) := Zero; end if;
    end loop;
    return E;
  end ToErr;

  procedure Prod(F1,F2: in Zernike; F3: in out Zernike) is
    LC1: constant Integer := LastC(F1.P,F1.C);
    LE1: constant Integer := LastE(F1.E);
    LC2,LE2: Integer;
  begin
    F3.R := RMin(F1.R,F2.R);
    F3.P := (F1.P+F2.P) mod 2;
    if (LC1<0) and then (LE1<0) then SetZero(F3); return; end if;
    LC2 := LastC(F2.P,F2.C);
    LE2 := LastE(F2.E);
    if (LC2<0) and then (LE2<0) then SetZero(F3); return; end if;
    if STrunc then
      Prod(F1.P,F2.P,F3.P,F1.C,F2.C,F3.C);
    else
      Prod(F1,F2,F3,LC1,LE1,LC2,LE2);
    end if;
  end Prod;

  procedure Prod(P1,P2,P3: in Parity; C1,C2: in Radials; C3: in out Radials) is
    --- assumes STrunc=True
    --- deterministic: tasks working on independent data
    CosSin: constant Boolean := (P1=0) and then (P2=1);
    SinCos: constant Boolean := (P1=1) and then (P2=0);
    SinSin: constant Boolean := (P1=1) and then (P2=1);
    W: constant RadVec(0..0) := (0 => Zero);
    NZ1: constant Boos := NonZero(P1,C1);
    NZ2: constant Boos := NonZero(P2,C2);
    Task_Error: Boolean := False;
    PC: Protected_Counter(P3);

    -- 2*cos(m)*cos(n) =  cos(m-n) + cos(m+n)
    -- 2*sin(m)*sin(n) =  cos(m-n) - cos(m+n)
    -- 2*cos(m)*sin(n) = -sin(m-n) + sin(m+n)
    -- 2*sin(m)*cos(n) =  sin(m-n) + sin(m+n)

    procedure Prod_Loop is
      M2,M3: Integer;
    begin
      Proper_Rounding;
      loop
        PC.Next(M3);
        exit when M3>Size;
        declare
          C3M3: Radial renames C3(M3);
        begin
          SetZero(M3,C3M3);
          for M1 in 0 .. M3/2 loop               -- freq sum
            M2 := M3-M1;
            if NZ2(M2) and then NZ1(M1) then
              IncludeProd(W,C1(M1),C2(M2),C3M3,SinSin);
            end if;
            if (M2/=M1) and then NZ2(M1) and then NZ1(M2) then
              IncludeProd(W,C1(M2),C2(M1),C3M3,SinSin);
            end if;
          end loop;
          for M1 in M3 .. Size loop       -- freq diff
            M2 := M1-M3;
            if NZ2(M2) and then NZ1(M1) then
              IncludeProd(W,C1(M1),C2(M2),C3M3,CosSin);
            end if;
            if (M3/=0) and then NZ2(M1) and then NZ1(M2) then
              IncludeProd(W,C1(M2),C2(M1),C3M3,SinCos);
            end if;
          end loop;
          Mult(SHalf,C3M3);
        end;
      end loop;
    exception
      when others => Task_Error := True; raise;
    end Prod_Loop;

    task type Prod_Task_Type is end Prod_Task_Type;
    task body Prod_Task_Type is begin Prod_Loop; end Prod_Task_Type;

    NT: constant Natural := Reserve_Tasks(Size,PProd_Parallel);
  begin
    if NT=0 then
      Prod_Loop;
    else
      declare
        Prod_Task: array(1 .. NT) of Prod_Task_Type;
      begin
        Prod_Loop;
      end;
      Free_Tasks(NT);
      if Task_Error then raise Sorry with "Task_Error"; end if;
    end if;
  end Prod;

  procedure Prod(F1,F2: in Zernike; F3: in out Zernike; LC1,LE1,LC2,LE2: in Integer) is
    --- assuming that STrunc is False, and that F3.R and F3.P have been set
    --- deterministic: tasks working on independent data
    P1: Parity renames F1.P;
    P2: Parity renames F2.P;
    P3: constant Parity := F3.P;
    R3: constant Radius := F3.R;
    CosSin: constant Boolean := (P1=0) and then (P2=1);
    SinCos: constant Boolean := (P1=1) and then (P2=0);
    SinSin: constant Boolean := (P1=1) and then (P2=1);
    CC: constant Boolean := (LC1>=0) and (LC2>=0);
    CE: constant Boolean := (LC1>=0) and (LE2>=0);
    EC: constant Boolean := (LE1>=0) and (LC2>=0);
    EE: constant Boolean := (LE1>=0) and (LE2>=0);
    LCC: constant Integer := LC1+LC2;
    LCE: constant Integer := LC1+LE2;
    LEC: constant Integer := LE1+LC2;
    LEE: constant Integer := LE1+LE2;
    MCE: constant Integer := IMin(LC1,LE2);
    MEC: constant Integer := IMin(LE1,LC2);
    MEE: constant Integer := IMin(LE1,LE2);
    LC3: constant Integer := IMin(LCC,Size);
    LE3: constant Integer := IMin(LEE,SizeE);
    Fac: constant Radius := (One/R3)/R3;
    W: constant WeightsE := NPowersE(R3);
    NZ1: constant Boos := NonZero(F1);
    NZ2: constant Boos := NonZero(F2);
    B1: constant Weights := ToErr(W,NZ1,F1);
    B2: constant Weights := ToErr(W,NZ2,F2);
    C1: Radials renames F1.C;
    C2: Radials renames F2.C;
    C3: Radials renames F3.C;
    E1: Errors renames F1.E;
    E2: Errors renames F2.E;
    E3: Errors renames F3.E;
    Task_Error: Boolean := False;
    C_PC: Protected_Counter(P3);
    E_PC: Protected_Counter(0);

    -- 2*cos(m)*cos(n) =  cos(m-n) + cos(m+n)
    -- 2*sin(m)*sin(n) =  cos(m-n) - cos(m+n)
    -- 2*cos(m)*sin(n) = -sin(m-n) + sin(m+n)
    -- 2*sin(m)*cos(n) =  sin(m-n) + sin(m+n)

    procedure AddCProd(M3: in Natural) is
      --- coeff*coeff
      --- divide by 2 later
      M2: Integer;
      C3M3: Radial renames C3(M3);
    begin
      for M1 in 0 .. M3/2 loop           -- freq sum
        M2 := M3-M1;
        if NZ2(M2) and then NZ1(M1) then
          IncludeProd(W,C1(M1),C2(M2),C3M3,SinSin);
        end if;
        if (M2/=M1) and then NZ2(M1) and then NZ1(M2) then
          IncludeProd(W,C1(M2),C2(M1),C3M3,SinSin);
        end if;
      end loop;
      for M1 in M3 .. Size loop          -- freq diff
        M2 := M1-M3;
        if NZ2(M2) and then NZ1(M1) then
          IncludeProd(W,C1(M1),C2(M2),C3M3,CosSin);
        end if;
        if (M3/=0) and then NZ2(M1) and then NZ1(M2) then
          IncludeProd(W,C1(M2),C2(M1),C3M3,SinCos);
        end if;
      end loop;
    end AddCProd;

    procedure AddEProd0 is
      --- terms that go into E3(0)
      --- divide by 2 later
      E: Flt renames E3(0);
      R,S: Flt;
    begin
      if EE then
        -------------------- err*err
        S := Zero;
        for M2 in reverse MEE+1 .. LE2 loop S := S+E2(M2); end loop;
        R := Zero;
        for M1 in reverse 0 .. MEE loop -- freq diff: M2 >= M1
          S := S+E2(M1);
          R := Fac*R+E1(M1)*S;
        end loop;
        E := E+R;
        S := Zero;
        for M1 in reverse MEE+1 .. LE1 loop S := S+E1(M1); end loop;
        R := Zero;
        for M2 in reverse 0 .. MEE loop -- freq diff: M1 > M2
          R := Fac*R+E2(M2)*S;
          S := S+E1(M2);
        end loop;
        E := E+R;
        E := E+E1(0)*E2(0);             -- freq sum
      end if;
      if CE then
        -------------------- coeff*err
        S := Zero;
        for M1 in reverse MCE+1 .. LC1 loop S := S+B1(M1); end loop;
        R := Zero;
        for M2 in reverse 0 .. MCE loop -- freq diff: M2 <= M1
          S := S+B1(M2);
          R := Fac*R+E2(M2)*S;
        end loop;
        E := E+R;
        E := E+B1(0)*E2(0);             -- freq sum
      end if;
      if EC then
        -------------------- err*coeff
        S := Zero;
        for M2 in reverse MEC+1 .. LC2 loop S := S+B2(M2); end loop;
        R := Zero;
        for M1 in reverse 0 .. MEC loop -- freq diff:  M1 <= M2
          S := S+B2(M1);
          R := Fac*R+E1(M1)*S;
        end loop;
        E := E+R;
        E := E+E1(0)*B2(0);             -- freq sum
      end if;
    end AddEProd0;

    procedure AddEProd(M3: in Positive) is
      --- divide by 2 later
      E: Radius renames E3(IMin(M3,SizeE));
      R: Flt;
    begin
      if EE then
        -------------------- err*err
        for M1 in IMax(0,M3-LE2) .. IMin(LE1,M3) loop
          E := E+E1(M1)*E2(M3-M1);             -- freq sum:  M1+M2=M3
        end loop;
      end if;
      if CE and then (M3<=LCE) then
        -------------------- coeff*err
        for M1 in IMax(0,M3-LE2) .. IMin(LC1,M3) loop
          E := E+B1(M1)*E2(M3-M1);             -- freq sum:  M1+M2=M3
        end loop;
        if M3<=LE2 then
          R := Zero;
          for M1 in 0 .. IMin(LC1,LE2-M3) loop -- freq diff: M2-M1=M3
            R := Fac*R+B1(M1)*E2(M1+M3);
          end loop;
          E := E+R;
        end if;
      end if;
      if EC and then (M3<=LEC) then
        -------------------- err*coeff
        for M2 in IMax(0,M3-LE1) .. IMin(LC2,M3) loop
          E := E+B2(M2)*E1(M3-M2);             -- freq sum:  M1+M2=M3
        end loop;
        if M3<=LE1 then
          R := Zero;
          for M2 in 0 .. IMin(LC2,LE1-M3) loop -- freq diff: M1-M2=M3
            R := Fac*R+B2(M2)*E1(M2+M3);
          end loop;
          E := E+R;
        end if;
      end if;
      if CC and then (M3>Size) and then (M3<=LCC) then
        -------------------- coeff*coeff
        for M1 in IMax(0,M3-LC2) .. IMin(LC1,M3) loop
          E := E+B1(M1)*B2(M3-M1);           -- freq sum:  M1+M2=M3
        end loop;
      end if;
    end AddEProd;

    procedure Prod_Loop is
      M3: Natural;
    begin
      Proper_Rounding;
      if CC then
        loop
          C_PC.Next(M3);
          exit when M3>LC3;
          AddCProd(M3);
        end loop;
      end if;
      loop
        E_PC.Next(M3);
        exit when M3>SizeE;
        if M3=0 then
          AddEProd0;
        else
          AddEProd(M3);
          if M3=SizeE then
            for M in SizeE+1 .. LEE loop
              AddEProd(M);
            end loop;
          end if;
        end if;
      end loop;
    exception
      when others => Task_Error := True; raise;
    end Prod_Loop;

    task type Prod_Task_Type is end Prod_Task_Type;
    task body Prod_Task_Type is begin Prod_Loop; end Prod_Task_Type;

    NT: constant Natural := Reserve_Tasks(IMax(LC3,LE3),PProd_Parallel);

  begin
    SetZero(F3);
    if NT=0 then
      Prod_Loop;
    else
      declare
        Prod_Task: array(1 .. NT) of Prod_Task_Type;
      begin
        Prod_Loop;
      end;
      Free_Tasks(NT);
      if Task_Error then raise Sorry with "Task_Error"; end if;
    end if;
    if P3=1 then Oddify(E3); end if;
    Mult(SHalf,F3);
  end Prod;

  function "*"(F1,F2: Zernike) return Zernike is
    F3: Zernike;
  begin
    Prod(F1,F2,F3);
    return F3;
  end "*";

  function MaxNorm(E: Errors) return Radius is
    R: Flt := E(0);
  begin
    for M in 1 .. SizeE loop R := R+E(M); end loop;
    return R;
  end MaxNorm;

  procedure Norm(F: in Zernike; S: in out Scalar) is
    R: Radius renames F.R;
    C: Radials renames F.C;
    Tmp: Scalar;
  begin
    if STrunc then SetZero(S); else BallAt0(MaxNorm(F.E),S); end if;
    for M in reverse F.P .. Size loop
      Norm(R,C(M),Tmp);
      Add(Tmp,S);
    end loop;
  end Norm;

  function Norm(F: Zernike) return Scalar is
    S: Scalar;
  begin
    Norm(F,S);
    return S;
  end Norm;

  function MaxNorm(F: Zernike) return Radius is
    W: constant Weights := NPowers(F.R);
    S: Radius;
    C: Radials renames F.C;
  begin
    if STrunc then S := Zero; else S := MaxNorm(F.E); end if;
    for M in reverse F.P .. Size loop
      S := S+MaxNorm(W,C(M));
    end loop;
    return S;
  end MaxNorm;

  procedure L2_Prod(F1,F2: in Zernike; S: in out Scalar) is
    --- assuming Not_STrunc and F1.P=F2.P
    W: constant WeightsE := NPowersE(One/RMin(F1.R,F2.R));
    C1: Radials renames F1.C;
    C2: Radials renames F2.C;
    E1: Errors renames F1.E;
    E2: Errors renames F2.E;
    SE1: Radius := E1(0);
    SE2: Radius := E2(0);
    E: Flt := Zero;
    T: Scalar;
  begin
    if F1.P=0 then
      L2Prod(SE1,SE2,W,C1(0),C2(0),S);
      Mult(Two,S);
    else
      SetZero(S);
    end if;
    for M in 1 .. Size loop       --- low*low
      SE1 := SE1+E1(M);
      SE2 := SE2+E2(M);
      L2Prod(SE1,SE2,W,C1(M),C2(M),T);
      Add(T,S);
    end loop;
    for M in Size+1 .. SizeE loop --- low*high + high*low + high*high
      SE1 := SE1+E1(M);
      E := E+(SE1*E2(M)+E1(M)*SE2)*(W(M)*W(M))/Flt(M+1);
      SE2 := SE2+E2(M);
    end loop;
    BallAt0(Half*E,T);
    Add(T,S);
  end L2_Prod;

  procedure L2Prod(F1,F2: in Zernike; S: in out Scalar) is
  begin
    SetZero(S);
    if F1.P=F2.P then
      if STrunc then
        declare
          W: constant RadVec(0..0) := (0 => Zero);
          C1: Radials renames F1.C;
          C2: Radials renames F2.C;
          T: Scalar;
        begin
          if F1.P=0 then
            L2Prod(W,C1(0),C2(0),S);
            Mult(Two,S);
          end if;
          for M in 1 .. Size loop
            L2Prod(W,C1(M),C2(M),T);
            Add(T,S);
          end loop;
        end;
      else
        L2_Prod(F1,F2,S);
      end if;
    end if;
  end L2Prod;

  procedure Coeff(M,I: in Natural; F: in Zernike; S: in out Scalar) is
    D: constant Integer := (Size-M)/2;
    P: Radial renames F.C(M);          -- error if M>Size
    PC: Poly renames P.L.all.Data;
  begin
    if (M<F.P) or (I>D) then raise Not_Implemented; end if;
    if STrunc then
      Copy(PC(I),S);
    else
      declare
        FE: Errors renames F.E;
        PE: Errs renames P.E;
        E: Radius := Zero;
      begin
        for K in 0 .. M loop
          E := E+FE(K);
        end loop;
        for K in 0 .. I loop
          E := E+PE(K);
        end loop;
        BallAt0(NPower(One/F.R,M+2*I)*E,S);
        Add(PC(I),S);
      end;
    end if;
  end Coeff;

  procedure AddCoeff(M,I: in Natural; S: in Scalar; F: in out Zernike) is
  begin
    if M<F.P then raise Undefined; end if;
    Add(S,F.C(M).L.all.Data(I));
  end AddCoeff;

  procedure AddHigher(M,Ie: in Natural; R: in Radius; F: in out Zernike) is
    --- modes (M,I) with I >= Ie
    E: Radius renames F.C(M).E(IMin((Size-M)/2+1,Ie));
  begin
    if M<F.P then raise Undefined; end if;
    if Not_STrunc then E := E+R; end if;
  end AddHigher;

  procedure AddHigher(Me: in Natural; R: in Radius; F: in out Zernike) is
    --- modes (M,I) with M >= Me and I >= 0
    E: Radius renames F.E(IMin(Size,IMax(F.P,Me)));
  begin
    if Not_STrunc then E := E+R; end if;
  end AddHigher;

  procedure AddBall(S: in Scalar; F: in out Zernike) is
  begin
    F.E(F.P) := F.E(F.P)+MaxNorm(S);
  end AddBall;

  procedure Show(N: in String; E: in Errors; Cut: in Flt := Zero) is
  begin
    for M in FreqE loop
      if E(M)>Cut then
        Show1(N & Strng(M,3) & "  ",E(M));
      end if;
    end loop;
  end Show;

  procedure Show(F: in Zernike; Cut: in Flt := Zero) is
    R: Radius renames F.R;
    C: Radials renames F.C;
  begin
    Show1("Cut =",Cut);
    Show1("Rho =",R);
    Show1("Parity =",F.P);
    Show0("Radials");
    for M in F.P .. Size loop
      if MaxNorm(R,C(M))>Cut then
        Show1(Strng(M,3) & "  ",C(M));
      end if;
    end loop;
    if Not_STrunc then
      Show0("Errors");
      Show("",F.E,Cut);
    end if;
  end Show;

  procedure Show1(N: in String; F: in Zernike; Hide0: in Boolean := True) is
    use RV;
  begin
    if not IsZero(F) then
      Show1(N & "R ",F.R);
      Show1(N & "P ",F.P);
      Show1(N & "C ",F.C(F.P .. Size),Hide0);
      if Not_STrunc then
        Show(N & "E ",F.E);
      end if;
    elsif not Hide0 then
      Show0(N & "0");
    end if;
  end Show1;

  procedure Put(F: in File_Type; H: in Zernike; Decimal: in Boolean := False) is
    use RV;
    C: Radials renames H.C;
    L: constant Natural := EffDeg(C);
  begin
    Put(F,H.R,Decimal);
    Put(F,H.P);
    Put(F,L);
    for I in H.P .. L loop Put(F,C(I),Decimal); end loop;
    Put(F,SizeE,H.E,Decimal);
  end Put;

  procedure Get(F: in File_Type; H: in out Zernike; Decimal: in Boolean := False) is
    R: Radius renames H.R;
    C: Radials renames H.C;
    E: Errors renames H.E;
    Cut: Boolean := False;
    L: Integer := 0;
    Me: Integer;
    B: Radius := Zero;
    P: Radial;
  begin
    SetZero(H);
    Get(F,R,Decimal);
    Get(F,H.P);
    Get(F,L);
    for M in H.P .. L loop
      P.M := M;
      Get(R,F,P,Decimal);
      if IsZero(P) then
        null;
      elsif M<=Size then
        Copy(P,C(M));
      elsif STrunc then
        Cut := True;
      else
        Me := IMin(M,SizeE);
        E(Me) := E(Me)+P.E(0);
      end if;
    end loop;
    Get(F,L);
    for M in 0 .. L loop
      Get(F,B,Decimal);
      if Not_STrunc then
        Me := IMin(SizeE,IMax(H.P,M));
        E(Me) := E(Me)+B;
      end if;
    end loop;
    if Cut or Decimal then
      Put_Line("Warning: Input truncated or inaccurate");
    end if;
  end Get;

  procedure Write(FileName: in String; H: in Zernike; Decimal: in Boolean := False) is
    F: File_Type;
  begin
    if Verbosity>0 then Show0("Writing " & FileName); end if;
    Create(F,Out_File,FileName);
    Put(F,H,Decimal);
    Close(F);
  end Write;

  procedure Read(FileName: in String; H: in out Zernike; Decimal: in Boolean := False) is
    F: File_Type;
  begin
    if Verbosity>0 then Show0("Reading " & FileName); end if;
    Open(F,In_File,FileName);
    Get(F,H,Decimal);
    Close(F);
  end Read;

  procedure Val(F: in Zernike; R,Theta: in Scalar; S: in out Scalar) is
    --- assumes R,Theta are real
    C: Radials renames F.C;
  begin
    SetZero(S);
    if F.P=0 then
      for M in 0 .. Size loop
        S := S+Val(C(M),R)*Cos(Scal(M)*Theta);
      end loop;
    else
      for M in 1 .. Size loop
        S := S+Val(C(M),R)*Sin(Scal(M)*Theta);
      end loop;
    end if;
    if Not_STrunc then
      S := S+BallAt0(MaxNorm(F.E));
    end if;
  end Val;

  function Val(F: Zernike; R,Theta: Scalar) return Scalar is
    S: Scalar;
  begin
    Val(F,R,Theta,S);
    return S;
  end Val;

  function Center0(F: Zernike) return Boolean is
    C: Radials renames F.C;
  begin
    for M in F.P .. Size loop
      if not Center0(C(M)) then return False; end if;
    end loop;
    return True;
  end Center0;

  procedure ErrMult(R: in Radius; F: in out Zernike) is
    C: Radials renames F.C;
    E: Errors renames F.E;
  begin
    if Not_STrunc then
      for M in F.P .. Size loop ErrMult(R,C(M)); end loop;
      for M in FreqE loop E(M) := R*E(M); end loop;
    end if;
  end ErrMult;

  function Contains(F1,F2: Zernike) return Logical is
  begin
    if STrunc then
      if F1=F2 then return True; else return False; end if;
    end if;
    if F2.P /= F1.P then return False; end if;
    if (F2.R<F1.R) and then (not IsSharp(F2)) then return False; end if;
    declare
      C1: Radials renames F1.C;
      C2: Radials renames F2.C;
      E1: Errors renames F1.E;
      E2: Errors renames F2.E;
      Answer: Logical := True;
    begin
      for M in F1.P .. Size loop
        Answer := Answer and Contains(C1(M),C2(M));
        if Answer=False then return False; end if;
      end loop;
      for M in FreqE loop
        if E1(M)<E2(M) then return False; end if;
      end loop;
      return Answer;
    end;
  end Contains;

  procedure NumTrunc(Siz: in Natural; F1: in Zernike; F2: in out Zernike) is
    C1: Radials renames F1.C;
    C2: Radials renames F2.C;
  begin
    F2.R := F1.R;
    F2.P := F1.P;
    for M in F1.P .. Size loop
      NumTrunc(Siz,C1(M),C2(M));
    end loop;
    if Not_STrunc then F2.E := (others => Zero); end if;
  end NumTrunc;

  procedure Raise_Errs(Siz: in Natural; F: in out Zernike) is
    --- raise errors above Siz
    R1: constant Radius := One/F.R;
    R2: constant Radius := R1*R1;
    FE: Errors renames F.E;
    FC: Radials renames F.C;
    RM,ESum: Flt;
  begin
    if Not_STrunc then
      if Siz>Size then raise Not_Implemented; end if;
      if F.P=0 then
        RM := One;
        ESum := Zero;
      else
        RM := R1;
        ESum := FE(0);
        FE(0) := Zero;
      end if;
      for M in F.P .. Siz loop
        ESum := ESum+FE(M);
        FE(M) := Zero;
        declare
          P: Radial renames FC(M);
        begin
          P.E(0) := P.E(0)+ESum;
          Raise_Errs(R2,RM,Siz,P);
        end;
        RM := RM*R1;
      end loop;
      FE(Siz+1) := FE(Siz+1)+ESum;
    end if;
  end Raise_Errs;

  procedure InvNegLap(F1: in Zernike; F2: in out Zernike) is
    P: Parity renames F1.P;
    Rho: Radius renames F1.R;
    C1: Radials renames F1.C;
    C2: Radials renames F2.C;
  begin
    F2.R := Rho;
    F2.P := P;
    for M in P .. Size loop
      InvNegLap(Rho,C1(M),C2(M));
    end loop;
    if Not_STrunc then
      declare
        Rho2: constant Radius := Rho*Rho;
        E2: Errors renames F2.E;
      begin
        E2 := F1.E;
        if P=1 then Oddify(E2); end if;
        for M in P .. SizeE loop
          E2(M) := InvLapNorm(Rho2,M)*E2(M);
        end loop;
      end;
    end if;
  end InvNegLap;

  function InvNegLap(F: Zernike) return Zernike is
    Ft: Zernike;
  begin
    InvNegLap(F,Ft);
    return Ft;
  end InvNegLap;

  procedure DerTheta(F1: in Zernike; F2: in out Zernike) is
    C2: Radials renames F2.C;
  begin
    if not IsSharp(F1) then raise Not_Implemented; end if;
    if F1.P=0 then
      Neg(F1,F2);
      F2.P := 1;
    else
      Copy(F1,F2);
      SetZero(0,C2(0));
      F2.P := 0;
    end if;
    for M in 2 .. Size loop Mult(Flt(M),C2(M)); end loop;
  end DerTheta;

  procedure RotatePi(F1: in Zernike; F2: in out Zernike) is
    C2: Radials renames F2.C;
  begin
    if not IsSharp(F1) or F1.P=1 then raise Not_Implemented; end if;
    Copy(F1,F2);
    for M in Freq loop
      if (M mod 2)=1 then Neg(C2(M)); end if;
    end loop;
  end RotatePi;

  procedure RotatePiHalf(F1: in Zernike; F2: in out Zernike) is
    C2: Radials renames F2.C;
  begin
    if not IsSharp(F1) or F1.P=1 then raise Not_Implemented; end if;
    Copy(F1,F2);
    for M in Freq loop
      if (M mod 2)=1 then
        SetZero(C2(M));
      elsif (M mod 4)=2 then
        Neg(C2(M));
      end if;
    end loop;
  end RotatePiHalf;

  procedure Simple_Random(R: in Radius; P: in Parity; F: in out Zernike) is
    C: Radials renames F.C;
  begin
    F.R := R;
    F.P := P;
    for M in P .. Size loop
      Simple_Random(R,M,C(M));
    end loop;
    if Not_STrunc then F.E := (others => Zero); end if;
  end Simple_Random;

  procedure Fill_One(R: in Radius; P: in Parity; F: in out Zernike) is
    --- for testing
    FC: Radials renames F.C;
    FE: Errors renames F.E;
  begin
    F.R := R;
    F.P := P;
    for M in P .. Size loop
      declare
        D: constant Integer := (Size-M)/2;
        P: Radial renames FC(M);
        PC: Poly renames P.L.all.Data;
        PE: Errs renames P.E;
      begin
        P.M := M;
        for I in 0 .. D loop Copy(SOne,PC(I)); end loop;
        if Not_STrunc then
          for I in 0 .. D+1 loop PE(I) := One; end loop;
        end if;
      end;
    end loop;
    if Not_Strunc then
      for M in 0 .. SizeE loop FE(M) := One; end loop;
    end if;
  end Fill_One;

  procedure Plot(F: in Zernike; L: Positive := 50) is
    use Flt_EF;
    P: constant Flt := 3.141592653589793;
    R,Theta: Flt;
    S: Scalar;
  begin
    for M in 1 .. L loop
      R := Flt(M)/Flt(L);
      for N in 0 .. L loop
        Theta := P*Flt(2*N)/Flt(L);
        S := Val(F,Scal(R),Scal(Theta));
        Show3("",R*Cos(Theta),R*Sin(Theta),Approx(S));
      end loop;
      New_Line;
    end loop;
  end Plot;

  procedure PutPlot(F: in File_Type; Z: in Zernike; L: in Positive := 50) is
    use FLt_IO,Flt_EF;
    Digs: constant Positive := 4;
    P: constant Flt := 3.141592653589793;
    R,Theta: Flt;
    S: Scalar;
  begin
    for M in 1 .. L loop
      R := Flt(M)/Flt(L);
      for N in 0 .. L loop
        Theta := P*Flt(2*N)/Flt(L);
        S := Val(Z,Scal(R),Scal(Theta));
        Put(F,R*Cos(Theta),3,Digs,4);
        Put(F,R*Sin(Theta),3,Digs,4);
        Put(F,Approx(S),3,Digs,4);
        New_Line(F);
      end loop;
    end loop;
  end PutPlot;

  function GuessRho(F: Zernike) return Radius is
    use Flt_EF;
    N0: constant Integer := Size/4;
    N1: constant Integer := (3*Size)/4;
    FC: Radials renames F.C;
    K,N: Integer := 0;
    T,R: Flt := One;
  begin
    for M in 0 .. N1 loop
      declare
        PC: Poly renames FC(M).L.all.Data;
      begin
        for I in IMax((N0-M)/2,0) .. (N1-M)/2 loop  --  N0 <=  M+2*I <=  N1
          N := M+2*I;
          if N /= 0 then
            T := MaxNorm(PC(I));
            if T /= Zero then
              K := K+1;
              R := R+Exp(Log(T)/Flt(N));
            end if;
          end if;
        end loop;
      end;
    end loop;
    R := R/Flt(K);
    return One/R;
  end GuessRho;

  procedure AddHill(S: in Scalar; F: in out Zernike) is
    C: Poly renames F.C(0).L.all.Data;
    Tmp: Scalar;
  begin
    if not IsZero(S) then
      if F.P=1 then raise Parity_Error; end if;
      AddProd( 0.05      ,S,C(0),Tmp);
      AddProd(-0.09      ,S,C(1),Tmp);
      AddProd( 0.1095    ,S,C(2),Tmp);
      AddProd(-0.107     ,S,C(3),Tmp);
      AddProd( 0.088875  ,S,C(4),Tmp);
      AddProd(-0.064515  ,S,C(5),Tmp);
      AddProd( 0.0416307 ,S,C(6),Tmp);
      AddProd(-0.0241686 ,S,C(7),Tmp);
      AddProd( 0.0127407 ,S,C(8),Tmp);
      AddProd(-0.00614497,S,C(9),Tmp);
    end if;
  end AddHill;

  procedure AddRing(S: in Scalar; F: in out Zernike) is
    C: Poly renames F.C(0).L.all.Data;
    Tmp: Scalar;
  begin
    if not IsZero(S) then
      if F.P=1 then raise Parity_Error; end if;
      AddProd( 0.546381   ,S,C(0),Tmp);
      AddProd( 0.310376   ,S,C(1),Tmp);
      AddProd(-0.263433   ,S,C(2),Tmp);
      AddProd(-0.179173   ,S,C(3),Tmp);
      AddProd( 0.0862047  ,S,C(4),Tmp);
      AddProd( 0.0446674  ,S,C(5),Tmp);
      AddProd(-0.0249888  ,S,C(6),Tmp);
      AddProd(-0.00466321 ,S,C(7),Tmp);
      AddProd( 0.00508884 ,S,C(8),Tmp);
      AddProd(-0.000430486,S,C(9),Tmp);
    end if;
  end AddRing;

  procedure AddPower(S: in Scalar; M: in Positive; F: in out Zernike) is
    C: Poly renames F.C(0).L.all.Data;
    O: constant Rational := 0/1;
    D: array(0..4) of Rational;
    Tmp: Scalar;
  begin
    if not IsZero(S) then
      if F.P=1 then raise Parity_Error; end if;
      case M is
        when 2 => D := (1/2,1/2,O,O,O);
        when 4 => D := (1/3,1/2,1/6,O,O);
        when 6 => D := (1/4,9/20,1/4,1/20,O);
        when 8 => D := (1/5,2/5,2/7,1/10,1/70);
        when others => raise Not_Implemented;
      end case;
      for I in D'Range loop
        AddProd(Scal(D(I)),S,C(I),Tmp);
      end loop;
    end if;
  end AddPower;

  procedure Normal(Lambda: in Scalar; XH,H: in out Zernike; Fac: in out Scalar) is
    --- use only in numeric mode
    S,T: Scalar;
  begin
    L2Prod(XH,H,S);
    Sqrt(S,T);
    Inv(T,S);
    Mult(S,XH);
    Mult(S,H);    -- now <XH,H>=1
    Neg(Lambda,Fac);
  end Normal;

  ---------------------------------------- modes

  procedure Norm(Q: in ZMode; S: in out Scalar) is
  begin
    if Q.Err then
      Copy(SOne,S);
    else
      declare
        R,Tmp: Scalar;
      begin
        Assign(Q.R,R);
        IPower(Q.M+2*Q.I,R,S,Tmp);
      end;
    end if;
  end Norm;

  procedure MaxNorm(Q: in ZMode; S: out Radius) is
  begin
    if Q.Err then
      S := One;
    else
      S := NPower(Q.R,Q.M+2*Q.I);
    end if;
  end MaxNorm;

  procedure Assign(Q: in ZMode; F: in out Zernike) is
    M: Natural renames Q.M;
    I: Natural renames Q.I;
  begin
    SetZero(Q.R,Q.P,F);
    if Q.Err then
      if I=0 then
        F.E(M) := One;
      else
        F.C(M).E(I) := One;
      end if;
    else
      Copy(SOne,F.C(M).L.all.Data(I));
    end if;
  end Assign;

  procedure AddProd(Q: in ZMode; S: in Scalar; F: in out Zernike; Check: in Boolean) is
    --- assuming Q is consistent
    M: Natural renames Q.M;
    I: Natural renames Q.I;
  begin
    if not IsZero(S) then
      if Check then
        if Q.P /= F.P then raise Parity_Error; end if;
        if Q.R>F.R then raise Domain_Error; end if;
      end if;
      if Q.Err then
        if I=0 then
          F.E(M) := F.E(M)+MaxNorm(S);
        else
          F.C(M).E(I) := F.C(M).E(I)+MaxNorm(S);
        end if;
      else
        Add(S,F.C(M).L.all.Data(I));
      end if;
    end if;
  end AddProd;

  procedure Extract_Coeff(Q: in ZMode; F: in out Zernike; S: in out Scalar) is
    --- Q.Err=False
    M: Natural renames Q.M;
    I: Natural renames Q.I;
    P: Radial renames F.C(M);
    PC: Poly renames P.L.all.Data;
  begin
    if Q.Simple then     -- move a coeff (into S)
      SetZero(S);
      Swap(PC(I),S);
    else                 -- move a coeff (into S) and shift errors
      declare
        PE: Errs renames P.E;
        FE: Errors renames F.E;
      begin
        if I=0 then
          if M=1 and then Q.P=1 then
            Oddify(F.E);
          end if;
          if FE(M) /= Zero then              -- split Zernike error
            PE(0) := PE(0)+FE(M);
            FE(M+1) := FE(M+1)+FE(M);
            FE(M) := Zero;
          end if;
        end if;
        if PE(I)=Zero then
          SetZero(S);
          Swap(PC(I),S);
        else                                 -- split Polar error
          BallAt0(PE(I),S);
          Add(PC(I),S);
          SetZero(PC(I));
          PE(I+1) := PE(I+1)+PE(I);
          PE(I) := Zero;
        end if;
      end;
    end if;
  end Extract_Coeff;

  procedure Extract_Zernike_Error(Q: in ZMode; F: in out Zernike; S: in out Scalar) is
    --- Q.Err=True and Q.I=0
    M: Natural renames Q.M;
    FE: Errors renames F.E;
  begin
    if Q.Simple then
      BallAt0(FE(M),S);
      FE(M) := Zero;
    else
      declare
        Rho: Radius renames F.R;
        FC: Radials renames F.C;
        N: Radius := Zero;
      begin
        for K in reverse M .. SizeE loop
          if FE(K) /= Zero then
            N := N+FE(K);
            FE(K) := Zero;
          end if;
        end loop;
        for K in reverse IMax(M,Q.P) .. Size loop
          if not IsZero(FC(K)) then
            N := N+MaxNorm(Rho,FC(K));
            SetZero(FC(K));
          end if;
        end loop;
        BallAt0(N,S);
      end;
    end if;
  end Extract_Zernike_Error;

  procedure Extract_Radial_Error(Q: in ZMode; F: in out Zernike; S: in out Scalar) is
    --- Q.Err=True and Q.I>0
    M: Natural renames Q.M;
    I: Natural renames Q.I;
    P: Radial renames F.C(M);
    PE: Errs renames P.E;
  begin
    if Q.Simple then
      BallAt0(PE(I),S);
      PE(I) := Zero;
    else                -- move modes (M,J) with J>=I  (into S)
      declare
        D: constant Natural := (Size-M)/2;
        Rho: Radius renames Q.R;
        Rho2: constant Radius := Rho*Rho;
        PC: Poly renames P.L.all.Data;
        N: Radius := Zero;
      begin
        for J in reverse I .. D loop
          N := Rho2*N;
          if not IsZero(PC(J)) then
            N := N+MaxNorm(PC(J));
            SetZero(PC(J));
          end if;
        end loop;
        if N>Zero then
          N := NPower(Rho,M+2*I)*N;
        end if;
        for J in reverse I .. D+1 loop
          if PE(J) /= Zero then
            N := N+PE(J);
            PE(J) := Zero;
          end if;
        end loop;
        BallAt0(N,S);
      end;
    end if;
  end Extract_Radial_Error;

  procedure Extract(Q: in ZMode; F: in out Zernike; S: in out Scalar; Check: in Boolean) is
    --- assuming Q is consistent
  begin
    if Check then
      if F.P /= Q.P then raise Parity_Error; end if;
      if Not_STrunc and then (Q.R>F.R) then raise Domain_Error; end if;
    end if;
    if Q.Err then
      if Q.I=0 then
        Extract_Zernike_Error(Q,F,S);
      else
        Extract_Radial_Error(Q,F,S);
      end if;
    else
      Extract_Coeff(Q,F,S);
    end if;
  end Extract;

  procedure Show(Q: in ZMode; NewLine: in Boolean := True) is
  begin
    if Q.Err then
      Show0("Err: P=" & Strng(Q.P,1) & " M=" & Strng(Q.M,3) & " I=" & Strng(Q.I,3),NewLine);
    else
      Show0("Coe: P=" & Strng(Q.P,1) & " M=" & Strng(Q.M,3) & " I=" & Strng(Q.I,3),NewLine);
    end if;
  end Show;

  procedure Show_More(Q: in ZMode; NewLine: in Boolean := True) is
  begin
    if Q.Simple then Show0("simple"); else Show0("not simple"); end if;
    Show1("R=",Q.R);
    Show(Q,NewLine);
  end Show_More;

  procedure Show(Q: in ZModes) is
    Q1: ZMode renames Q(1);
  begin
    Show_More(Q1);
    for J in Q'Range loop
      declare
        QJ: ZMode renames Q(J);
      begin
        Show0(Strng(J,4) & " ",False);
        Show(QJ,False);
        if not ((QJ.Simple=Q1.Simple) and (QJ.P=Q1.P) and (QJ.R=Q1.R)) then
          Show0(" !!! wrong entry");
        end if;
        Show0;
      end;
    end loop;
  end Show;

  procedure Check(Q: in ZMode) is
    M: Natural renames Q.M;
    I: Natural renames Q.I;
  begin
    if STrunc and then (Q.Err or (not Q.Simple)) then
      Show0("STrunc Err or not Simple");
      raise Inconsistent_Modes;
    end if;
    if M<Q.P then
      Show0("M < Parity");
      raise Inconsistent_Modes;
    end if;
    if Q.Err then
      if I=0 then
        if M <= SizeE then return; end if;
        Show1("Zernike Err M ",M);
        raise Inconsistent_Modes;
      end if;
      if (M+2*I)>(Size+2) then
        Show2("Radial Err M I ",M,I);
        raise Inconsistent_Modes;
      end if;
    elsif (M+2*I)>Size then
      Show2("Coeff M I ",M,I);
      raise Inconsistent_Modes;
    end if;
  end Check;

  procedure Check_Order(Q1,Q2: in ZMode) is
    --- for adjacent modes
    procedure Bail_Out is
    begin
      Show0("!!! inconsistent pair:");
      Show_More(Q1);
      Show_More(Q2);
      raise Inconsistent_Modes;
    end Bail_Out;
  begin
    if Q2.M < Q1.M then Bail_Out; end if;
    if (Q2.M = Q1.M) and then (Q2.I <= Q1.I) then Bail_Out; end if;
    if (Q2.M > Q1.M) then
      if (not Q1.Err) then Bail_Out; end if;
      if Q2.Err and then (Q2.I /= 0) then Bail_Out; end if;
    end if;
    if (Q2.M = Q1.M) and then Q1.Err then Bail_Out; end if;
    if Q1.Err and then (Q1.I = 0) then Bail_Out; end if;
  end Check_Order;

  procedure Check_Basics(Q: in ZModes) is
    F: constant Positive := Q'First;
    QF: ZMode renames Q(F);
    R: Radius renames QF.R;
    P: Parity renames QF.P;
    Simple: Boolean renames QF.Simple;
  begin
    if F /= 1 then raise Inconsistent_Modes; end if;
    Check(QF);
    for N in 2 .. Q'Last loop
      declare
        QN: ZMode renames Q(N);
      begin
        Check(QN);
        if not ((QN.R=R) and (QN.P=P) and (QN.Simple=Simple)) then
          Show_More(QN);
          raise Inconsistent_Modes;
        end if;
      end;
    end loop;
  end Check_Basics;

  procedure Check(Q: in ZModes) is
    QF: ZMode renames Q(Q'First);
  begin
    Check_Basics(Q);
    if not QF.Simple then
      declare
        QL: ZMode renames Q(Q'Last);
        S: Scalar;
        F: Zernike;
      begin
        for N in 2 .. Q'Last loop Check_Order(Q(N-1),Q(N)); end loop;
        if not (QL.Err and QL.I=0) then
          raise Inconsistent_Modes;
        end if;
        Fill_One(QF.R,QF.P,F);
        for N in Q'Range loop Extract(Q(N),F,S,False); end loop;
        if not IsZero(F) then
          Show0("Q is incomplete");
          Show1("",F);
        end if;
      end;
    end if;
  end Check;

  function Make(Siz: Positive; Rho: Radius; P: Parity; Simple,Trunc: in Boolean) return ZModes is
    Include_Errs: constant Boolean := not (Trunc or STrunc);
    Simp: constant Boolean := Simple or Trunc or STrunc;

    procedure Build(Q: out ZModes; Count: out Natural) is
      procedure Set_Mode(M,I: in Natural; Err: in Boolean) is
      begin
        Count := Count+1;
        if Count <= Q'Last then Q(Count) := (Rho,P,M,I,Err,Simp); end if;
      end Set_Mode;
      D: Natural;
    begin
      Count := 0;
      for M in P .. Siz loop
        D := (Siz-M)/2;
        for I in 0 .. D loop
          Set_Mode(M,I,False);
        end loop;
        if Include_Errs then
          Set_Mode(M,D+1,True);
        end if;
      end loop;
      if Include_Errs then
        Set_Mode(Siz+1,0,True);
      end if;
    end Build;

    Count: Natural;
  begin
    declare
      Q: ZModes(1..1);
    begin
      Build(Q,Count);
    end;
    declare
      Q: ZModes(1..Count);
    begin
      Build(Q,Count);
      return Q;
    end;
  end Make;

end Zernikes;
