with Ada.Text_IO, Logicals, Scalars, Ints, Rationals, Flts;
use Ada.Text_IO, Logicals, Scalars, Ints, Rationals, Flts;
with Pooling, Vectors, Regges;

pragma Elaborate_All (Ada.Text_IO,Logicals,Scalars,Ints,Rationals,Flts);
pragma Elaborate_All (Pooling,Vectors,Regges);

generic

  Size: in Positive;

  -------------------- begin_include scalars Scalar

  type Scalar is private;

  --- basic
  with function Info(Dummy: Scalar) return Scalar_Info is <>;
  with function IsSharp(S: Scalar) return Boolean is <>;
  with function IsZero(S: Scalar) return Boolean is <>;
  with function "="(S1,S2: Scalar) return Boolean is <>;
  with procedure SetZero(S: in out Scalar) is <>;
  with procedure Copy(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Swap(S1,S2: in out Scalar) is <>;

  --- sets
  with function Center0(S: Scalar) return Boolean is <>;
  with function Contains0(S: Scalar) return Logical is <>;
  with function Contains(S1,S2: Scalar) return Logical is <>;
  with procedure BallAt0(R: in Flt; S: in out Scalar) is <>;
  with function BallAt0(R: Flt) return Scalar is <>;
  with procedure ToErr(S: in out Scalar) is <>;
  with procedure ToErr(S1: in Scalar; S2: in out Scalar) is <>;
  with function ToErr(S: Scalar) return Scalar is <>;
  with procedure Center(S: in out Scalar) is <>;
  with procedure Center(S1: in Scalar; S2: in out Scalar) is <>;
  with function Center(S: Scalar) return Scalar is <>;
  with procedure ModCenter(S: in out Scalar) is <>;
  with procedure ModCenter(S1: in Scalar; S2: in out Scalar) is <>;
  with function ModCenter(S: Scalar) return Scalar is <>;
  with procedure ErrMult(R: in Radius; S: in out Scalar) is <>;
  with procedure Union(S1: in Scalar; S2: in out Scalar) is <>;
  with function Union(S1,S2: Scalar) return Scalar is <>;
  with procedure Intersection(S1: in Scalar; S2: in out Scalar; Empty: out Logical) is <>;

  --- order
  with function Sign(S: Scalar) return Integer is <>;
  with function Compare(S1,S2: Scalar) return Integer is <>;
  with function "<"(S1,S2: Scalar) return Boolean is <>;
  with function "<="(S1,S2: Scalar) return Boolean is <>;
  with function ">"(S1,S2: Scalar) return Boolean is <>;
  with function ">="(S1,S2: Scalar) return Boolean is <>;
  with procedure Min(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Min(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with function Min(S1,S2: Scalar) return Scalar is <>;
  with procedure Max(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Max(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with function Max(S1,S2: Scalar) return Scalar is <>;
  with function Inf(S: Scalar) return Flt is <>; -- inf of real part
  with function Sup(S: Scalar) return Flt is <>; -- sup of real part

  --- arithmetic
  with procedure Neg(S: in out Scalar) is <>;
  with procedure Neg(S1: in Scalar; S2: in out Scalar) is <>;
  with function "-"(S: Scalar) return Scalar is <>;
  with procedure Add(I: in Integer; S: in out Scalar) is <>;
  with procedure Add(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Sum(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with function "+"(S1,S2: Scalar) return Scalar is <>;
  with procedure Sub(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Diff(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with function "-"(S1,S2: Scalar) return Scalar is <>;
  with procedure Mult(R: in Flt; S,Tmp: in out Scalar) is <>;
  with procedure Mult(R: in Flt; S: in out Scalar) is <>;
  with procedure Prod(R: in Flt; S1: in Scalar; S2: in out Scalar) is <>;
  with function "*"(R: Flt; S: Scalar) return Scalar is <>;
  with procedure AddProd(R: in Flt; S1: in Scalar; S2,Tmp: in out Scalar) is <>;
  with procedure AddProd(R: in Flt; S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Mult(Q: in Rational; S: in out Scalar) is <>;
  with procedure Mult(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Prod(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with function "*"(S1,S2: Scalar) return Scalar is <>;
  with procedure AddProd(S1,S2: in Scalar; S3,Tmp: in out Scalar) is <>;
  with procedure AddProd(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with procedure SubProd(S1,S2: in Scalar; S3,Tmp: in out Scalar) is <>;
  with procedure Div(R: in Flt; S,Tmp: in out Scalar) is <>;
  with procedure Div(R: in Flt; S: in out Scalar) is <>;
  with procedure Quot(S1: in Scalar; R: in Flt; S2: in out Scalar) is <>;
  with function "/"(S: Scalar; R: Flt) return Scalar is <>;
  with procedure Div(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Quot(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with function "/"(S1,S2: Scalar) return Scalar is <>;
  with procedure Inv(S1: in Scalar; S2: in out Scalar) is <>;
  with function Inv(S: Scalar) return Scalar is <>;
  with function "**"(S: Scalar; I: Integer) return Scalar is <>;

  --- fun
  with function IsReal(S: Scalar) return Boolean is <>;
  with procedure Adjoint(S: in out Scalar) is <>;
  with procedure Adjoint(S1: in Scalar; S2: in out Scalar) is <>;
  with function Adjoint(S: Scalar) return Scalar is <>;
  with procedure Real_Part(S: in out Scalar) is <>;
  with procedure Imag_Part(S: in out Scalar) is <>;
  with procedure Eval0(S: in out Scalar) is <>;
  with function "abs"(S: Scalar) return Scalar is <>;
  with procedure Norm(S1: in Scalar; S2: in out Scalar) is <>;
  with function Norm(S: Scalar) return Scalar is <>;
  with function MaxNorm(S: Scalar) return Radius is <>;
  with procedure Sqr(S: in out Scalar) is <>;
  with function Sqr(S: Scalar) return Scalar is <>;
  with procedure IPower(I: in Integer; S1: in Scalar; S2,Tmp: in out Scalar) is <>;
  with procedure Sqrt(S1: in Scalar; S2: in out Scalar) is <>;
  with function Sqrt(S: Scalar) return Scalar is <>;
  with procedure Root(K: in Positive; S1: in Scalar; S2: in out Scalar) is <>;
  with function Root(K: Positive; S: Scalar) return Scalar is <>;
  with procedure Exp(S1: in Scalar; S2: in out Scalar) is <>;
  with function Exp(S: Scalar) return Scalar is <>;
  with procedure ArcCos(S1: in Scalar; S2: in out Scalar) is <>;
  with function ArcCos(S: Scalar) return Scalar is <>;
  with procedure ArcSin(S1: in Scalar; S2: in out Scalar) is <>;
  with function ArcSin(S: Scalar) return Scalar is <>;
  with procedure Cos(S1: in Scalar; S2: in out Scalar) is <>;
  with function Cos(S: Scalar) return Scalar is <>;
  with procedure Sin(S1: in Scalar; S2: in out Scalar) is <>;
  with function Sin(S: Scalar) return Scalar is <>;
  with procedure Log(S1: in Scalar; S2: in out Scalar) is <>;
  with function Log(S: Scalar) return Scalar is <>;
  with procedure Simple_Random(S: in out Scalar) is <>;
  with procedure QPower(Q: in Rational; S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Roots2(B,C: in Scalar; U1,U2,V: in out Scalar) is <>;
  with procedure Roots3(B,C,D: in Scalar; U0,U1,U2,V: in out Scalar) is <>;

  --- conversion and i/o
  with function Approx(S: Scalar) return Flt is <>;
  with procedure Assign(K: in Integer; S: in out Scalar) is <>;
  with procedure Assign(Q: in Rational; S: in out Scalar) is <>;
  with procedure Assign(R: in Flt; S: in out Scalar) is <>;
  with function Scal(I: Integer) return Scalar is <>;
  with function Scal(Q: Rational) return Scalar is <>;
  with function Scal(R: Flt) return Scalar is <>;
  with procedure Enclose(R1,R2: in Flt; S: in out Scalar) is <>;
  with function Enclose(R1,R2: Flt) return Scalar is <>;
  with procedure Show1(N: in String; S: in Scalar; Whatever: in Boolean := True) is <>;
  with procedure Show2(N: in String; S1,S2: in Scalar; Whatever: in Boolean := True) is <>;
  with procedure Put(F: in File_Type; S: in Scalar; Decimal: in Boolean := False) is <>;
  with procedure Get(F: in File_Type; S: in out Scalar; Decimal: in Boolean := False) is <>;

  -------------------- end_include scalars

  with procedure Proper_Rounding is <>;
  with procedure CGS(I: in Positive; S: in out Scalar) is <>; -- squares of CG coefficients

package Zernikes is

  PDeg: constant Positive := Size/2;
  PDeg1: constant Positive := PDeg+1;
  type Poly is array(0 .. PDeg) of Scalar;
  subtype Errs is RadVec(0 .. PDeg1);

  function FirstE(Last: Natural; E: in Errs) return Natural;
  function LastE(Last: Natural; E: Errs) return Integer;
  function LastC(Last: Natural; C: Poly) return Integer;
  function EffDeg(Last: Natural; C: Poly) return Natural;
  procedure CP(C1: in Poly; C2: in out Poly);

  package RP is new Pooling (Item => Poly);
  subtype Pooled_Poly is RP.Controlled_LD_Pointer;

  type Radial is new Pooled_Poly with
    --- L: LD_Pointer := null;      -- L.Data(i) is the coefficient of R^m_n with n=|m|+2*i
    record
      M: Natural;                   -- angular frequency
      E: Errs;                      -- higher order error norms
    end record;

  --- vector component operations
  function "="(P1,P2: Radial) return Boolean;
  function IsZero(P: Radial) return Boolean;
  function IsReal(P: Radial) return Boolean;
  procedure SetZero(P: in out Radial);
  procedure Copy(P1: in Radial; P2: in out Radial);
  procedure Swap(P1,P2: in out Radial);
  procedure Neg(P: in out Radial);
  procedure Neg(P1: in Radial; P2: in out Radial);
  procedure Add(P1: in Radial; P2: in out Radial);
  procedure Sum(P1,P2: in Radial; P3: in out Radial);
  procedure Sub(P1: in Radial; P2: in out Radial);
  procedure Diff(P1,P2: in Radial; P3: in out Radial);
  procedure Real_Part(P: in out Radial);
  procedure Imag_Part(P: in out Radial);
  procedure Adjoint(P: in out Radial);
  procedure Adjoint(P1: in Radial; P2: in out Radial);
  procedure Show1(N: in String; P: in Radial; Hide0: in Boolean := True);
  procedure Show2(N: in String; P1,P2: in Radial; Hide0: in Boolean := True);
  procedure Put(F: in File_Type; P: in Radial; Decimal: in Boolean := False);     -- does not save P.M
  procedure Get(F: in File_Type; P: in out Radial; Decimal: in Boolean := False); -- not implemented

  --- other simple operations
  procedure SetZero(M: in Natural; P: in out Radial);
  function IsSharp(P: Radial) return Boolean;         --- no errors
  procedure Mult(R: in Flt; P: in out Radial);
  procedure Prod(R: in Flt; P1: in Radial; P2: in out Radial);
  procedure AddProd(R: in Flt; P1: in Radial; P2: in out Radial);
  procedure Mult(S: in Scalar; P: in out Radial);
  procedure Prod(S: in Scalar; P1: in Radial; P2: in out Radial);
  procedure AddProd(S: in Scalar; P1: in Radial; P2: in out Radial);
  procedure SubProd(S: in Scalar; P1: in Radial; P2: in out Radial);
  procedure Div(R: in Flt; P: in out Radial);
  procedure Quot(P1: in Radial; R: in Flt; P2: in out Radial);
  function Val(P: Radial; X: Scalar) return Scalar;
  function Center0(P: Radial) return Boolean;
  procedure ErrMult(R: in Radius; P: in out Radial);
  function Contains(P1,P2: Radial) return Logical;
  procedure NumTrunc(Siz: in Natural; P1: in Radial; P2: in out Radial);

  --- more
  procedure Get(Rho: in Radius; F: in File_Type; P: in out Radial; Decimal: in Boolean := False); -- set P.M beforehand
  procedure Norm(Rho: in Radius; P: in Radial; S: in out Scalar);
  function MaxNorm(Rho: Radius; P: Radial) return Radius;
  function MaxNorm(W: RadVec; P: Radial) return Radius;
  procedure L2Prod(SE1,SE2: in Radius; W: in RadVec; P1,P2: in Radial; S: in out Scalar);
  procedure L2Prod(W: in RadVec; P1,P2: in Radial; S: in out Scalar);
  procedure AddProd(W: in RadVec; P1,P2: in Radial; P3: in out Radial);
  procedure IncludeProd(W: in RadVec; P1,P2: in Radial; P3: in out Radial; Subtract: in Boolean);
  procedure Raise_Errs(R2,RM: in Radius; Siz: in Natural; P: in out Radial);
  procedure InvNegLap(Rho: in Radius; P1: in Radial; P2: in out Radial); -- inverse of -Laplacean
  function InvLapNorm(Rho2: in Radius; M: Natural) return Radius;
  procedure Simple_Random(R: in Radius; M: in Natural; P: in out Radial);

  --- now Zernike

  package RV is new Vectors (Component => Radial);

  SizeE:  constant Positive := 2*Size;

  subtype Freq  is Integer range 0 .. Size;
  subtype FreqE is Integer range 0 .. SizeE;
  subtype Radials is RV.Vector(Freq);
  subtype Errors is RadVec(FreqE);
  type Boos is array(Freq) of Boolean;

  type Zernike is
    record
      R: Radius;     -- determines domain of analyticity
      P: Parity;     -- 0=even, 1=odd
      C: Radials;    -- coefficients of cos(m .) or sin(m.)
      E: Errors;     -- errors
    end record;

  function "="(F1,F2: Zernike) return Boolean;
  procedure SetZero(F: in out Zernike);
  procedure SetZero(R: in Radius; F: in out Zernike);
  procedure SetZero(R: in Radius; P: in Parity; F: in out Zernike);
  function IsZero(F: Zernike) return Boolean;
  function IsSharp(F: Zernike) return Boolean;
  function NonZero(F: Zernike) return Boos;
  procedure Copy(F1: in Zernike; F2: in out Zernike);
  procedure Swap(F1,F2: in out Zernike);
  procedure Neg(F: in out Zernike);
  procedure Neg(F1: in Zernike; F2: in out Zernike);
  function "-"(F: Zernike) return Zernike;
  procedure Add(S: in Scalar; F: in out Zernike);
  procedure Add(F1: in Zernike; F2: in out Zernike);
  procedure Sum(F1,F2: in Zernike; F3: in out Zernike);
  function "+"(F1,F2: Zernike) return Zernike;
  procedure Sub(F1: in Zernike; F2: in out Zernike);
  procedure Diff(F1,F2: in Zernike; F3: in out Zernike);
  function "-"(F1,F2: Zernike) return Zernike;
  procedure Mult(S: in Scalar; F: in out Zernike);
  procedure Prod(S: in Scalar; F1: in Zernike; F2: in out Zernike);
  function "*"(S: Scalar; F: Zernike) return Zernike;
  procedure AddProd(S: in Scalar; F1: in Zernike; F2: in out Zernike);
  procedure Mult(F1: in Zernike; F2: in out Zernike);
  procedure Prod(F1,F2: in Zernike; F3: in out Zernike);
  function "*"(F1,F2: Zernike) return Zernike;
  procedure Norm(F: in Zernike; S: in out Scalar);
  function Norm(F: Zernike) return Scalar;
  function MaxNorm(F: Zernike) return Radius;
  procedure L2Prod(F1,F2: in Zernike; S: in out Scalar);

  procedure Coeff(M,I: in Natural; F: in Zernike; S: in out Scalar);
  procedure AddCoeff(M,I: in Natural; S: in Scalar; F: in out Zernike);
  procedure AddHigher(M,Ie: in Natural; R: in Radius; F: in out Zernike);
  procedure AddHigher(Me: in Natural; R: in Radius; F: in out Zernike);
  procedure AddBall(S: in Scalar; F: in out Zernike);
  procedure Show(F: in Zernike; Cut: in Flt := Zero);
  procedure Show1(N: in String; F: in Zernike; Hide0: in Boolean := True);
  procedure Put(F: in File_Type; H: in Zernike; Decimal: in Boolean := False);
  procedure Get(F: in File_Type; H: in out Zernike; Decimal: in Boolean := False);
  procedure Write(FileName: in String; H: in Zernike; Decimal: in Boolean := False);
  procedure Read(FileName: in String; H: in out Zernike; Decimal: in Boolean := False);
  procedure Val(F: in Zernike; R,Theta: in Scalar; S: in out Scalar);
  function Val(F: Zernike; R,Theta: Scalar) return Scalar;
  function Center0(F: Zernike) return Boolean;
  procedure ErrMult(R: in Radius; F: in out Zernike);
  function Contains(F1,F2: Zernike) return Logical;
  procedure NumTrunc(Siz: in Natural; F1: in Zernike; F2: in out Zernike);
  procedure Raise_Errs(Siz: in Natural; F: in out Zernike);
  procedure InvNegLap(F1: in Zernike; F2: in out Zernike);
  function InvNegLap(F: Zernike) return Zernike;
  procedure DerTheta(F1: in Zernike; F2: in out Zernike);
  procedure RotatePi(F1: in Zernike; F2: in out Zernike);
  procedure RotatePiHalf(F1: in Zernike; F2: in out Zernike);
  procedure Simple_Random(R: in Radius; P: in Parity; F: in out Zernike);
  procedure Fill_One(R: in Radius; P: in Parity; F: in out Zernike);
  procedure Plot(F: in Zernike; L: Positive := 50);
  procedure PutPlot(F: in File_Type; Z: in Zernike; L: in Positive := 50);
  function GuessRho(F: Zernike) return Radius;

  procedure AddHill(S: in Scalar; F: in out Zernike);
  procedure AddRing(S: in Scalar; F: in out Zernike);
  procedure AddPower(S: in Scalar; M: in Positive; F: in out Zernike);

  --- modes etc

  type ZMode is  -- a unit coeff or unit error
    record
      R: Radius;
      P: Parity;
      M: Natural;
      I: Natural;
      Err: Boolean;    -- if True: I=0 means Zernike error, I>0 means Polar error
      Simple: Boolean; -- for maps that do not mess with error components
    end record;

  procedure Norm(Q: in ZMode; S: in out Scalar);
  procedure MaxNorm(Q: in ZMode; S: out Radius);
  procedure Assign(Q: in ZMode; F: in out Zernike);
  procedure AddProd(Q: in ZMode; S: in Scalar; F: in out Zernike; Check: in Boolean);
  procedure Extract(Q: in ZMode; F: in out Zernike; S: in out Scalar; Check: in Boolean);
  procedure Show(Q: in ZMode; NewLine: in Boolean := True);
  procedure Check(Q: in ZMode);

  type ZModes is array (Positive range <>) of ZMode; -- should start with index 1

  procedure Show(Q: in ZModes);
  procedure Check_Basics(Q: in ZModes);
  procedure Check(Q: in ZModes);
  function Make(Siz: Positive; Rho: Radius; P: Parity; Simple,Trunc: in Boolean) return ZModes;

  Inconsistent_Modes: exception;
  PProd_Parallel: Boolean := True;

  package Reggez is new Regges (LMax => Size);

private

  SZero:    constant Scalar      := Scal(0);
  SOne:     constant Scalar      := Scal(1);
  SThree:   constant Scalar      := Scal(Three);
  SHalf:    constant Scalar      := Scal(Half);
  SNegHalf: constant Scalar      := Scal(NegHalf);
  SInfo:    constant Scalar_Info := Info(SZero);
  STrunc:   constant Boolean     := SInfo.IsNumeric;
  Not_STrunc:  constant Boolean := not STrunc;

  subtype Weights is RadVec(Freq);
  subtype WeightsE is RadVec(FreqE);

  procedure Put(F: in File_Type; D: in Natural; E: in RadVec; Decimal: in Boolean := False);
  function NPower(R: Radius; N: Natural) return Radius;
  function NPowers(Rho: Radius) return Weights;
  procedure Fix_Signs(M1,M2: in out Integer; M3: in Integer);
  procedure AddProdCC(W: in RadVec; M1,M2,M3,D1,D2: in Natural; C1,C2: in Poly; C3: in out Poly; E3: in out Radius);
  procedure AddProdCE(W: in RadVec; M1,M2,M3,L1,L2: in Natural; C1: in Poly; E2: in Errs; E3: in out Errs);
  procedure InvNegLap(Rho: in Radius; M: in Natural; C1: in Poly; C2: in out Poly; E2: out Radius);

  function MaxNorm(E: Errors) return Radius;
  procedure Show(N: in String; E: in Errors; Cut: in Flt := Zero);

  function NonZero(P: Parity; C: Radials) return Boos;
  procedure Oddify(E: in out Errors);
  function LastE(E: Errors) return Integer;
  function LastC(First: Parity; C: Radials) return Integer;
  function ToErr(W: RadVec; NZ: Boos; F: Zernike) return Weights;
  procedure Prod(P1,P2,P3: in Parity; C1,C2: in Radials; C3: in out Radials);
  procedure Prod(F1,F2: in Zernike; F3: in out Zernike; LC1,LE1,LC2,LE2: in Integer);
  procedure L2_Prod(F1,F2: in Zernike; S: in out Scalar);
  procedure Normal(Lambda: in Scalar; XH,H: in out Zernike; Fac: in out Scalar);

  procedure Extract_Coeff(Q: in ZMode; F: in out Zernike; S: in out Scalar);
  procedure Extract_Zernike_Error(Q: in ZMode; F: in out Zernike; S: in out Scalar);
  procedure Extract_Radial_Error(Q: in ZMode; F: in out Zernike; S: in out Scalar);
  procedure Show_More(Q: in ZMode; NewLine: in Boolean := True);
  procedure Check_Order(Q1,Q2: in ZMode);

end Zernikes;
